% This program was used in publication "Ndc80 complex, a conserved coupler for kinetochore-microtubule motility, 
% is a sliding molecular clutch" by V. Demidov et al. 2025.
% This program automatically finds gliding interaction segments based on the indicated velocity threshold and combines them into gliding interaction events.
% Input: 
% 1. folder with QPD1 and QPD2 data files, each containing two columns: trapped bead position in nm and trap position in nm. 
% 2. Threshold defined after using Build_and_analyze_vel_hist.m program

clear all


%% INITAL PARAMETERS %%
down = 0; % Direction of plus-end gliding, 0 for up direction, 1 for down direction 
thr = 60;% threshold value, nm/s;
RangeOfRepeats=[1:999999]; % Gliding interaction segments consisting of this number of sweeps will be analyzed. Keep last value
dur_filtration = 0.002; % threshold for interaction segments
osc_range = 175; % Amplitude of oscillation
IndexShift =0; % 0 for 4-8 pN 3 for 12-20 pN
dt=15*1e-6; %time frame
grey = [0.3 0.3 0.3];
evdurthr = 0.002; % gliding segment duration threshold (code line: 514)


%% load data
pattern = ['*__0_61*']; % pattern to analyze 
listing =dir(pattern);
if listing(1).name(3)~=char('1')
    error('wrong file loaded')
elseif listing(2).name(3)~='2'
    error('wrong file loaded')
end
%load files
QD1=load(listing(1).name);
QD2=load(listing(2).name);


%% generate main array of data: A (position; force)

clear data
% check for file size
if size(QD1,1)>2002500
    error('wrong file size')
end

if down == 1
data(:,1) = QD1(:,1);
else
data(:,1) = QD2(:,1);
end

close all
clear A
sz=size(data);
if sz(2)==1
    poscol = 1;
else
    poscol = 2;
end

A(:,1) = data(:,poscol);% name(2000:end,poscol);
A(:,2) = (QD1(:,2)-QD1(:,1)+QD2(:,2)-QD2(:,1))./2;% name(2000:end,poscol+1)-name(2000:end,poscol);
clear name
clear UpIndex index;
clear DownIndex;

workpath=pwd;%uigetdir(pwd,'save files to');
filename='';
sz=size (A);
W0fast=8; %Gaussian filter window 1
W0slow=40; %Gaussian filter window 2

% indices of all force transitions
Force_multiplicated = A(1:end-1,2).*A(2:end,2);
index = find(Force_multiplicated<0);
index = index+IndexShift;

%hotfix:takes out first sweep if it is down direction
if sign(A(1,2))<0
A(1:index,:)=[];
index = index-index(1);
end

%% Find data frames when applied force switches direction and separate periods of up- and down-directed motion

time=0:dt:(size(A,1)-1)*dt; % time axis
UpIndex = [];
DownIndex = [];
UpIndex (:,1) = index(1:2:length(index))+1;
DownIndex (:,1) = index(2:2:length(index))+1;
if mod(length(index),2) == 0
    DownIndex (:,2) = [index(3:2:length(index)); size(A,1)];
    UpIndex (:,2) = [index(2:2:length(index))];    
else
    DownIndex (:,2) = [index(3:2:length(index))];
    UpIndex (:,2) = [index(2:2:length(index)); size(A,1)];    
end

%% Concatenate Up & Down motion into two separate time series data sets
clear Up Down;
Ashift = A(:,1);
for i=1:1:max(size(UpIndex))-1
    shiftUp = Ashift(UpIndex(i,2),1)-Ashift(UpIndex(i+1,1),1);
    regionUp = UpIndex(i+1,1):UpIndex(i+1,2);
    Ashift(regionUp,1) = Ashift(regionUp,1)+shiftUp;
    
end 

for i=1:1:max(size(DownIndex))-1
    regionDown = DownIndex(i+1,1):DownIndex(i+1,2);
    shiftDown = Ashift(DownIndex(i,2),1)-Ashift(DownIndex(i+1,1),1);
    Ashift(regionDown,1) = Ashift(regionDown,1)+shiftDown;
end

Up = zeros(size(A,1),1);
Upreal = zeros(size(A,1),1);
Downreal = zeros(size(A,1),1);
lastframe = 0;

for i=1:1:max(size(UpIndex))
    regionUp = UpIndex(i,1):UpIndex(i,2);
    framelength = UpIndex(i,2)-UpIndex(i,1)+1;
    Up (lastframe+1:lastframe+framelength) = Ashift(regionUp,1);
    Upreal (lastframe+1:lastframe+framelength) = A(regionUp,1);
    lastframe = lastframe+framelength;
end
    Up(lastframe+1:end) = [];
    Upreal(lastframe+1:end) = [];
Down = zeros(size(A,1),1);

lastframe = 0;
for i=1:1:max(size(DownIndex))
    regionDown = DownIndex(i,1):DownIndex(i,2);
    framelength = DownIndex(i,2)-DownIndex(i,1)+1;
    Down (lastframe+1:lastframe+framelength) = Ashift(regionDown,1);
    Downreal (lastframe+1:lastframe+framelength) = A(regionDown,1);
    lastframe = lastframe+framelength;
end
    Down(lastframe+1:end) = [];
    Downreal(lastframe+1:end) = [];

%% Find point-by-point derivative
clear derUp derUpsm derDown derDownsm
for i=1:1:size(Up)-1
    derUp(i)=1e-3*(Up(i+1)-Up(i))/dt; %um/s
end
for i=1:1:size(Down)-1
    derDown(i)=1e-3*(Down(i+1)-Down(i))/dt; %um/s
end

% gaussian filtering
derUpsmfast=smoothts(derUp,'g',8*W0fast,W0fast^2);
derUpsmslow=smoothts(derUp,'g',8*W0slow,W0slow^2);

%% Only one dirction is analyzed further, it has all prefixes as for Up signal for simplicity

if down == 1
    derUpsmslow=-smoothts(derDown,'g',8*W0slow,W0slow^2);
derUpsmfast=-smoothts(derDown,'g',8*W0fast,W0fast^2);  
derUp = derDown;
Upreal=Downreal;
UpIndex=DownIndex;
Up = Down;
direction='down\';
else
    direction='up\';
end

%% Plot velocity distribution, and fit it with gaussian

figure(1)
plot(time(1:length(derUpsmfast)),derUpsmfast,'color',grey)
hold on 
plot(time(1:length(derUpsmfast)),derUpsmslow,'black','Linewidth',1.5)
title('Velocity-vs-time')
xlabel('time, s'); ylabel('velocity, um/s');

figure(2)
binsz = 2;
hist(derUpsmslow,min(derUpsmslow):binsz:max(derUpsmslow))  %   derUpsmfast
hold on
%
[histy,histx] = hist(derUpsmslow,min(derUpsmslow):binsz:max(derUpsmslow));
[pks,locs,w,p]=findpeaks(histy,'MinPeakWidth',5); % small peaks usually have width ~ 2-3 points, right peaks have ~20-30 ponts
[peak_max_ind,ind_peak] = max(locs);
peak_width = w(ind_peak);


plot(histx(locs), w,'.r')

if peak_max_ind-peak_width<1
    peak_start = 1;
else
    peak_start = peak_max_ind-peak_width;
end
    
histx = histx(peak_start:peak_max_ind+0.4*peak_width); % 20 poits - approximate width of free vel peak
histy = histy(peak_start:peak_max_ind+0.4*peak_width); % 20 poits - approximate width of free vel peak
fitting2 = fit (histx', histy', 'gauss1'); % use [a,b,c]=coeffvalues(fitting1) to get fitted values
coeff = coeffvalues(fitting2);
plot(fitting2)
%
plot([thr thr], [0 0.5e3],'r')
title(['Vel distribution; mean vel = ', num2str(coeff(2),3),' +- ',num2str(coeff(3),3),'um/s, '])
xlim([-0.5e1 3.5e2])
hold off

%% searching for events by fast derivative<thr 

diff_vel = (derUpsmfast-derUpsmslow)';
diff_vel_mult = diff_vel(2:end).*diff_vel(1:end-1);
cross_ind = find(diff_vel_mult<=0); % indexes at when velocities are crossing
cross_ind = [1;cross_ind; length(diff_vel_mult)];
cross_ind (derUpsmslow(cross_ind(:,1))<=thr) = [];

k=1;
eventind = [];
for i=1:1:size(cross_ind,1)-1
    if min(derUpsmfast(cross_ind(i,1)+1:cross_ind(i+1,1)-1))>=thr
    else
        eventind(k,:) = [cross_ind(i,1),cross_ind(i+1,1)];
        k=k+1;
    end
end

i=1;
while i~=size(eventind,1)-1
    if eventind(i,2)==eventind(i+1,1)
        eventind(i,2) = eventind(i+1,2);
        eventind(i+1,:)=[];
    else
    i=i+1;    
    end
end

figure(1)
for i=1:1:size(eventind,1)
    t_loc = eventind(i,1):eventind(i,2);
    vel_loc = derUpsmslow(t_loc);
    plot(dt*t_loc,vel_loc,'LineWidth', 3)
end
plot(dt*[1 size(derUpsmslow,2)],[thr thr])

clear temp transindUp

%% taking out events that last > evdurthr
takeoutind=[];
for i=1:1:size(eventind,1)
    if dt*(eventind(i,2)-eventind(i,1))<evdurthr
        takeoutind(end+1) = i;
    end
end
eventind(takeoutind,:)=[];

%% finding indices of switchpoints within concatenated position data set
temp=UpIndex(:,2)-UpIndex(:,1)+1;                           
hold off

for i=1:length(temp)
transindUp(i) = sum(temp(1:i));% indices of breaks of Updersm
end

%% breaking gliding segments that last more then one sweep
clear temp
n=1;
takeoutind=[];
mainadd=[];
for i=1:size(eventind,1)
   temp=transindUp(:)-eventind(i,1);
n = find(temp==min(temp(temp>0))); 
    if eventind(i,2)-eventind(i,1)-temp(n)>0        % finding events that are longer then one sweep. Transindup(n) - index of beginning of the sweep that lies between start and end of gliding segment
        k=0;
        add=[];
            while eventind(i,2)>transindUp(n+k)     % finding k - how many sweeps are lying withing one event
                k=k+1;
            end
            if k==1
                add(1,1) = eventind(i,1);
                add(1,2) = transindUp(n);
                add(2,1) = transindUp(n)+1;
                add(2,2) = eventind(i,2);
                add(:,3) = i;
            else
                for j=1:k+1
                    if j==1                             
                        add(j,1) = eventind(i,1);
                        add(j,2) = transindUp(n);
                        add(:,3) = i;
                        continue
                    elseif j==k+1
                        add(j,1) = transindUp(n+(j-2))+1;
                        add(j,2) = eventind(i,2);
                        add(:,3) = i;
                        continue
                    else
                        add(j,1) = transindUp(n+j-2)+1;         % making a block of split events
                        add(j,2) = transindUp(n+j-1);
                        add(:,3) = i;
                    end
                end
            end
        mainadd = [mainadd;add];                                % making array of split events mainadd
    end
end

%% Specifying indexes of gliding segments according to the sweep edges

if isempty(mainadd)
else
addvect = mainadd(2:end,3)-mainadd(1:end-1,3);

sz = size(mainadd);
blockini = find(addvect~=0);      
blockini = [1;blockini+1];          % raw indexes where separate blocks start
blockl = [blockini(2:end)-blockini(1:end-1)-1;sz(1)-blockini(end)];

for i=1:size(blockini)
    blockend=blockini(i)+blockl(i);
    if i==1
        rawtorep = mainadd(blockini(i),3); % raw in eventind to replace
    else
        rawtorep = mainadd(blockini(i),3)+sum(blockl(1:i-1));
    end
    eventind(rawtorep,:)=[];
    eventind = [eventind(1:rawtorep-1,:) ; mainadd(blockini(i):blockend,1:2) ; eventind(rawtorep:end,:)];
end
end


% sort events in time
[~,idx] = sort(eventind(:,1)); % sort just the first column
eventind = eventind(idx,:);   % sort the whole matrix using the sort indices



%% Building array with index of each gliding segment and index of sweep it belongs to
k=1;
eventsweep=[];
combind=[];
for i=1:size(eventind,1)
    while eventind(i,1) >= transindUp(k)
        k=k+1;
    end
    if k==1
        shift = eventind(i,1);    % shift to dislpace ith event index when plotting on up+down graph
        eventsweep=[eventsweep; shift, i ,1];
    else
    shift = eventind(i,1)-transindUp(k-1)+1;    % shift to dislpace ith event from kth sweep beginning index when plotting on up+down graph
    if shift<0
            error('shift<0')
    end           

    eventsweep = [eventsweep; shift, i ,k];       % matrix of shift, event and sweep index
    end


end

eventindreal=[];

for i=1:size(eventind,1)
    ind=eventsweep(i,3);
    eventindreal(i,:) = eventind(i,:)-eventind(i,1)+eventsweep(i,1)+UpIndex(ind,1);
    
end

figure (7)
plot(time,A(:,1))
hold on

%% combining events that overlap if any

todel=[];
intnumber=0;
for i=1:size(eventindreal,1)-1
    if eventindreal(i,2) == eventindreal(i+1,2)
        eventindreal(i+1,1) = eventindreal(i,1);
        todel(end+1)=i;
    else
        if (dt*(eventindreal(i,2)-eventindreal(i,1)))<dur_filtration
            plot(time(eventindreal(i,1):eventindreal(i,2)),A(eventindreal(i,1):eventindreal(i,2)),'black')
        else
            plot(time(eventindreal(i,1):eventindreal(i,2)),A(eventindreal(i,1):eventindreal(i,2)),'r')
            intnumber = intnumber+1;
        end
    end
end

eventindreal(todel,:)=[];
eventind(todel,:)=[];
eventsweep(todel,:)=[];

hold off

%% Plotting detected gliding segments on the time series trapped bead coordinate data
figure(3)
    hold on
    t = dt*[1:1:length(Upreal)];
    plot(t,Upreal)
    bt_slope_std=[];
% calculation of slopes, sd and binding time
for i=1:1:size(eventind,1)
            deriv = (Upreal(eventind(i,1):eventind(i,2)));
            coeffs = polyfit(time(eventind(i,1):eventind(i,2)), deriv', 1);
            x = 1:1:length(deriv);
            x = x*dt+time(eventind(i,1));
            bt_slope_std(end+1,2) = coeffs(1,1)/1e3;
         bt_slope_std(end,1) = dt*(eventind(i,2)-eventind(i,1));
         bt_slope_std(end,3)=std(deriv-polyval(coeffs,x)');
%          plot(time(eventind(i,1):eventind(i,2)),polyval(coeffs, time(eventind(i,1):eventind(i,2))),'g')
        plot(t(eventind(i,1):eventind(i,2)),Upreal(eventind(i,1):eventind(i,2)),'r')
end

hold off

bt = (eventind(:,2)-eventind(:,1))*dt;
C=[];
for i=1:size(bt_slope_std,1)
    if bt_slope_std(i,1)>=dur_filtration
        C(end+1,:) = bt_slope_std(i,:);
    end
end

%% Saving data

directory=exist([workpath,'\',filename,direction,'graphs'],'dir');
if directory == 0
    mkdir ([workpath,'\',filename,direction,'graphs'])
    save([workpath,'\',filename,direction,'eventind.txt'], 'eventind', '-ascii', '-double', '-tabs')
    save([workpath,'\',filename,direction,'eventindreal.txt'], 'eventindreal', '-ascii', '-double', '-tabs')
    save([workpath,'\',filename,direction,'shift_events_sweep.txt'], 'eventsweep', '-ascii', '-double', '-tabs')
    save([workpath,'\',filename,direction,'UpIndex.txt'], 'UpIndex', '-ascii', '-double', '-tabs')
    save([workpath,'\',filename,direction,'DownIndex.txt'], 'DownIndex', '-ascii', '-double', '-tabs')
    save([workpath,'\',filename,direction,'bt_slope_std_filt.txt'], 'C', '-ascii', '-double', '-tabs')
    save([workpath,'\',filename,direction,'bt_slope_std.txt'], 'bt_slope_std', '-ascii', '-double', '-tabs')

    saveas(figure(1),[workpath,'\',filename,direction,'graphs\velocity filtered'])
    saveas(figure(2),[workpath,'\',filename,direction,'graphs\velocity distr'])
    saveas(figure(3),[workpath,'\',filename,direction,'graphs\bead position'])
    saveas(figure(7),[workpath,'\',filename,direction,'graphs\both dir bead position'])
    


else %reply=input('rewrite the graphs? y/n? ','s')
    save([workpath,'\',filename,direction,'eventind.txt'], 'eventind', '-ascii', '-double', '-tabs')
    save([workpath,'\',filename,direction,'eventindreal.txt'], 'eventindreal', '-ascii', '-double', '-tabs')
    save([workpath,'\',filename,direction,'shift_events_sweep.txt'], 'eventsweep', '-ascii', '-double', '-tabs')
    save([workpath,'\',filename,direction,'UpIndex.txt'], 'UpIndex', '-ascii', '-double', '-tabs')
    save([workpath,'\',filename,direction,'DownIndex.txt'], 'DownIndex', '-ascii', '-double', '-tabs')
    save([workpath,'\',filename,direction,'bt_slope_std_filt.txt'], 'C', '-ascii', '-double', '-tabs')
    
    save([workpath,'\',filename,direction,'bt_slope_std.txt'], 'bt_slope_std', '-ascii', '-double', '-tabs')
    saveas(figure(1),[workpath,'\',filename,direction,'graphs\velocity filtered'])
    saveas(figure(2),[workpath,'\',filename,direction,'graphs\velocity distr'])
    saveas(figure(3),[workpath,'\',filename,direction,'graphs\bead position'])
    saveas(figure(7),[workpath,'\',filename,direction,'graphs\both dir bead position'])
end


close all

