"""Module to estimate friction coefficient of the microtubule and compute forces from signals.

Conventions: forces are in pN, distances are in um, time in s, other units are derived from these."""

from pathlib import Path
from pprint import pformat
from typing import Tuple

import numpy as np
import pandas as pd
from scipy.optimize import fsolve


class Experiment:
    """Class to store the results of the experiment or simulation."""

    def __init__(self, **kwargs) -> None:
        self.__dict__.update(kwargs)

    def __repr__(self) -> str:
        return pformat(self.as_dict())

    def as_dict(self) -> dict:
        return self.__dict__

    @staticmethod
    def t(x: float) -> float:
        """Force-extension dependency of the dumbbell.

        Obtained from fitting a piecewise function to experimental data.

        :param x: extension of the dumbbell in um
        :return: force in pN
        """
        kw = 11.523647583728927
        wb = 0.08301813028722799
        a = 197.9917179948737
        b = -21.350156896852127
        c = 1.3645608917028682
        sb = 0.15757336810080158
        ks = 41.04628682418055
        os = -3.551448005614747

        if x < 0:
            return 0
        elif 0 <= x < wb:
            return kw * x
        elif wb <= x < sb:
            return a * x**2 + b * x + c
        else:
            return ks * x + os

    def prestretch_calc(self, y: np.ndarray) -> list:
        """Residuals of prestretch equations."""
        d1, d2 = y
        return [self.t(2 * d1) - self.fpre, self.t(2 * d2) - self.fpre]

    def solve_prestretch(self) -> list:
        """Determines initial deformations of MT springs after prestretch."""
        self.d1, self.d2 = fsolve(self.prestretch_calc, [0, 0])
        return self.d1, self.d2

    def _leading_bead(self, f2: float) -> float:
        """Force balance for the leading trap."""
        return f2 - self.fpre - self.f / 2

    def _trailing_bead(self, f1: float, delta1: float, delta2: float) -> float:
        """Force balance for the trailing trap."""
        return (
            f1
            - self.fpre
            + self.k1 / self.k2 * self.f / 2
            - self.k1 * (self.d1 + self.d2 - delta1 - delta2)
        )

    def _leading_dumbbell_spring(self, f1: float, delta1: float, v: float) -> float:
        """Force balance for the leading bead."""
        return self.t(2 * delta1) - f1 - self.g * v

    def _trailing_dumbbell_spring(self, f2: float, delta2: float, v: float) -> float:
        """Force balance for the trailing bead."""
        return self.t(2 * delta2) - f2 + self.g * v

    def _molecule(
        self, f1: float, f2: float, fmol: float, v: float, gmt: float
    ) -> float:
        """Force balance for the molecule."""
        return fmol - f2 + f1 + (2 * self.g + gmt) * v

    def eqns_free(self, x: np.ndarray) -> list[float, ...]:
        """Equations for the free motion of the dumbbell.

        :param x: list of variables in the system of equations (delta1, delta2, f1, f2, gmt)
        :return: residuals
        """
        delta1, delta2, f1, f2, gmt = x
        return [
            self._leading_bead(f2),
            self._trailing_bead(f1, delta1, delta2),
            self._leading_dumbbell_spring(f1, delta1, self.v_free),
            self._trailing_dumbbell_spring(f2, delta2, self.v_free),
            self._molecule(f1, f2, 0, self.v_free, gmt),
        ]

    def eqns_interaction(self, x: np.ndarray) -> list[float, ...]:
        """Equations for the intracting dumbbell motion.

        :param x: list of variables in the system of equations (delta1, delta2, f1, f2, fmol)
        :return: residuals
        """
        delta1, delta2, f1, f2, fmol = x
        return [
            self._leading_bead(f2),
            self._trailing_bead(f1, delta1, delta2),
            self._leading_dumbbell_spring(f1, delta1, self.v),
            self._trailing_dumbbell_spring(f2, delta2, self.v),
            self._molecule(f1, f2, fmol, self.v, self.gmt),
        ]

    def eqns_force_vel(self, x: np.ndarray) -> list[float, ...]:
        """Equations for the interacting dumbbell with given force-velocity
        dependency of the molecule.

        Requires ``self.force_vel`` method. It should accept force on the molecule
        and returns corresponding velocity

        :param x: list of variables in the system of equations (delta1, delta2, f1, f2, fmol)
        :return: residuals
        """
        delta1, delta2, f1, f2, fmol = x
        v = self.force_vel(fmol)
        return [
            self._leading_bead(f2),
            self._trailing_bead(f1, delta1, delta2),
            self._leading_dumbbell_spring(f1, delta1, v),
            self._trailing_dumbbell_spring(f2, delta2, v),
            self._molecule(f1, f2, fmol, v, self.gmt),
        ]

    def check_trap_distance(
        self, delta1: float, delta2: float, f1: float, f2: float
    ) -> bool:
        """Checks whether the distance between traps is the same as it was during prestretch."""
        prestretch = self.d1 + self.d2 + self.fpre / self.k1 + self.fpre / self.k2
        current = delta1 + delta2 + f1 / self.k1 + f2 / self.k2
        return np.allclose(prestretch, current)

    def solve_free(
        self, verbose: bool = False, fsolve_kwargs: dict | None = None
    ) -> tuple:
        """Solves system of equations if there is no dumbbell-molecule interaction."""
        if fsolve_kwargs is None:
            fsolve_kwargs = dict(factor=0.1)

        xsol, opt_res, code, msg = fsolve(
            self.eqns_free,
            [0, 0, self.fpre - self.f / 2, self.fpre + self.f / 2, 0.02],
            full_output=True,
            **fsolve_kwargs,
        )
        if verbose and code != 1:
            print(self)
            print("Warning message")
            print(xsol)
            print(opt_res)
            print(code)
            print(msg)

        res_arr = opt_res["fvec"]
        if not np.allclose(res_arr, np.zeros_like(res_arr)):
            print(f"Method did not achieve default numpy accuracy, residuals {res_arr}")

        self.gmt = xsol[-1]
        return xsol, opt_res, code, msg

    def solve_interacting(
        self, verbose: bool = False, fsolve_kwargs: dict | None = None
    ) -> tuple:
        """Solves system of equations if there is dumbbell-molecule interaction."""
        if fsolve_kwargs is None:
            fsolve_kwargs = dict(
                factor=0.1,
                xtol=1e-07,
                maxfev=1000,
            )
        xsol, opt_res, code, msg = fsolve(
            self.eqns_interaction,
            [0, 0, self.fpre - self.f / 2, self.fpre + self.f / 2, 0],
            full_output=True,
            **fsolve_kwargs,
        )
        if verbose and code != 1:
            print(self)
            print("Warning message")
            print(xsol)
            print(opt_res)
            print(code)
            print(msg)

        res_arr = opt_res["fvec"]
        if not np.allclose(res_arr, np.zeros_like(res_arr)):
            print(f"Method did not achieve default numpy accuracy, residuals {res_arr}")

        if not self.check_trap_distance(*xsol[:4]):
            raise RuntimeError(f"Constant trap distance violated!\n{self}")

        self.ext1 = xsol[0]
        self.ext2 = xsol[1]
        self.f1 = xsol[2]
        self.f2 = xsol[3]
        self.fmol = xsol[-1]
        return xsol, opt_res, code, msg

    def solve_from_force_vel(
        self, verbose: bool = False, fsolve_kwargs: dict | None = None
    ) -> tuple:
        """Solves the system of equations for interacting dumbbell
        assuming given force-velocity dependency for the molecule.
        """
        if fsolve_kwargs is None:
            fsolve_kwargs = dict(
                factor=0.1,
                xtol=1e-07,
                maxfev=1000,
            )
        xsol, opt_res, code, msg = fsolve(
            self.eqns_force_vel,
            [0, 0, self.fpre - self.f / 2, self.fpre + self.f / 2, 0],
            full_output=True,
            **fsolve_kwargs,
        )
        if verbose and code != 1:
            print(self)
            print("Warning message")
            print(xsol)
            print(opt_res)
            print(code)
            print(msg)

        res_arr = opt_res["fvec"]
        if not np.allclose(res_arr, np.zeros_like(res_arr)):
            print(f"Method did not achieve default numpy accuracy, residuals {res_arr}")

        if not self.check_trap_distance(*xsol[:4]):
            raise RuntimeError(f"Constant trap distance violated!\n{self}")

        self.ext1 = xsol[0]
        self.ext2 = xsol[1]
        self.f1 = xsol[2]
        self.f2 = xsol[3]
        self.fmol = xsol[-1]
        self.v = self.force_vel(self.fmol)
        return xsol, opt_res, code, msg


def pipeline(data: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Processes experimental data.
    Returns corrected datasets with forces and other estimated parameters.
    """
    sel_data_R = data[["exp_name", "F", "v_free_R", "v_R", "k_R", "k_L"]].rename(
        columns={"F": "f", "v_free_R": "v_free", "v_R": "v", "k_R": "k2", "k_L": "k1"}
    )

    sel_data_L = data[["exp_name", "F", "v_free_L", "v_L", "k_R", "k_L"]].rename(
        columns={"F": "f", "v_free_L": "v_free", "v_L": "v", "k_R": "k1", "k_L": "k2"}
    )

    transformed_dfs = dict()
    for side, sel_data in zip(["R", "L"], [sel_data_R, sel_data_L]):
        res_list = []

        for t in sel_data.itertuples():
            params = t._asdict()
            params["fpre"] = 2
            # All experimental data was obtained under 2 pN prestretch
            params["g"] = 0.0052  # Friction coefficient of a dumbbell bead

            ex = Experiment(**params)
            ex.solve_prestretch()
            ex.solve_free()
            ex.solve_interacting()
            res_list.append(ex.as_dict())

        transformed_dfs[side] = (
            pd.DataFrame(res_list).set_index("Index").rename_axis(None, axis=0)
        )

    res = pd.merge(
        transformed_dfs["R"],
        transformed_dfs["L"].drop(["g", "f", "fpre", "k1", "k2", "d1", "d2"], axis=1),
        how="outer",
        on="exp_name",
        suffixes=("_R", "_L"),
    )

    # swap f1 and f2 to denote Left and Right ends of the system, not leading/trailing ends
    res.rename(columns={"f2_L": "f1_L", "f1_L": "f2_L"}, inplace=True)

    # all forces are positive if force vector is parallel to direction (faces the same way)
    res["ftot_L"] = -(res["f2_L"] - res["f1_L"])
    res["ftot_R"] = res["f2_R"] - res["f1_R"]
    # res['fmol_L'] = -res['fmol_L']    -- no need to negate Fmol

    corr_res = res.rename(
        columns={
            "fmol_R": "Fmol_corr_R",
            "fmol_L": "Fmol_corr_L",
            "ftot_R": "Ftot_corr_R",
            "ftot_L": "Ftot_corr_L",
            "gmt_R": "gamma_MT_R",
            "gmt_L": "gamma_MT_L",
        }
    )[
        [
            "exp_name",
            "Fmol_corr_R",
            "Fmol_corr_L",
            "Ftot_corr_R",
            "Ftot_corr_L",
            "gamma_MT_R",
            "gamma_MT_L",
        ]
    ]

    return res, corr_res


def main(
    input_file: str | Path, all_results_file: str | Path, corr_results_file: str | Path
) -> None:
    """Runs the adjustment pipeline."""

    data = pd.read_excel(input_file)
    res, corr_res = pipeline(data)

    res.to_csv(all_results_file, index=False)
    corr_res.to_csv(corr_results_file, index=False)


if __name__ == "__main__":
    # root_folder = Path(...)
    #
    # input_file = root_folder / "delta80.xlsx"
    # all_results_file = root_folder / "2021-05-21_delta80_adjusted_full.csv"
    # corr_results_file = root_folder / "2021-05-21_delta80_corrected.csv"

    main(input_file, all_results_file, corr_results_file)
