"""Module with tools to load datasets."""

import os
from typing import Dict, Optional, Collection

import numpy as np
import pandas as pd
from uffc.utils import (
    get_contigious_ranges,
    inverse_dict_bijection,
    to_narrow,
    add_asymmetry_narrow,
    prepare_nonlinear_cols,
)

default_cols_to_negate: list[str] = ["G", "G_rot", "avg_dG_R", "avg_dG_L"]
default_abs_cols: list[str] = ["Ftot", "Fmol_L", "Fmol_R"]


# LEGACY function for obsolete data format
def get_exp_data_legacy(path: str, narrow: bool = True) -> pd.DataFrame:
    """Get experimental data. LEGACY function for OBSOLETE data format.

    :param path: path to CSV file with experimental results
    :param narrow: if True, convert to narrow format (labels for direction, positive F and v)
    :return: prepared DataFrame
    """

    df = pd.read_csv(path, header=None).rename(columns={0: "Ftot", 1: "v"})
    if narrow:
        df["direction"] = np.where(df["v"] >= 0, "R", "L")
        df.loc[:, ["Ftot", "v"]] = df.loc[:, ["Ftot", "v"]].applymap(abs)
    return df


def read_sim_table(path: str) -> pd.DataFrame:
    """Get columns from file with simulations list and their parameters.

    :param path: path to the file with simulations records
    :return: resulting DataFrame
    """
    cols = ["N", "B", "prestretchTotalForce", "trapstiffL", "trapstiffR"]
    df = pd.read_excel(path, usecols=cols).rename(
        columns={"prestretchTotalForce": "Fpre", "trapstiffL": "k1", "trapstiffR": "k2"}
    )
    df["Fpre"] /= 2.0
    return df


def set_negated_cols(
    df: pd.DataFrame, cols_to_negate: Optional[Collection[str]] = None
) -> pd.DataFrame:
    """Negate specified columns.

    :param df: input DataFrame
    :param cols_to_negate: if not None, then negate specified columns
    :return: new DataFrame with negated columns
    """
    tmp = df.copy()
    if cols_to_negate is not None:
        tmp.loc[:, cols_to_negate] = -tmp.loc[:, cols_to_negate]
    return tmp


def set_abs_cols(
    df: pd.DataFrame, abs_cols: Optional[Collection[str]] = None
) -> pd.DataFrame:
    """Take absolute values in given files.

    :param df: input DataFrame
    :param abs_cols: if not None, then take absolute values in specified columns
    :return: new DataFrame with abs-ed columns
    """
    tmp = df.copy()
    if abs_cols is not None:
        tmp.loc[:, abs_cols] = tmp.loc[:, abs_cols].abs()
    return tmp


def get_sim_data(
    path: str,
    cols_to_negate: Optional[Collection[str]] = None,
    abs_cols: Optional[Collection[str]] = None,
) -> pd.DataFrame:
    """Get simulation results data.

    :param path: path to file with simulation results and processed metrics
    :param cols_to_negate: if not None, then negate specified columns
    :param abs_cols: if not None, then replace values in specified columns with their absolute values
    """
    rename_map = inverse_dict_bijection(
        {
            "exp_num": "#",
            "G": "Gdepth (kT)",
            "Ftot": "Set Ftot (pN)",
            "Fmol": "Fmol - dir R (pN)",
            "Fmol_sd": "SD Fmol - dir R (pN)",
            "Fmol_N": "N (Fmol R)",
            "v_R": " Velocity - bead R - dir R (um/s)",
            "v_R_sd": "SD  Velocity - bead R - dir R (um/s)",
            "v_R_N": "N (Vel r R)",
            "v_L": " Velocity - bead R - dir L (um/s)",
            "v_L_sd": "SD  Velocity - bead R - dir L (um/s)",
            "v_L_N": "N (Vel r L)",
        }
    )

    tmp = pd.read_csv(path)
    tmp.rename(columns=rename_map, inplace=True)

    tmp = set_negated_cols(tmp, cols_to_negate)
    tmp = set_abs_cols(tmp, abs_cols)
    return tmp


def get_joined_data(
    root_folder: str,
    sim_table: str,
    data_table: str,
    cols_to_negate: Optional[Collection[str]] = None,
    abs_cols: Optional[Collection[str]] = None,
) -> pd.DataFrame:
    """Get data with additional stats.

    :param root_folder: folder that contains both table with stat and table with simulation results
    :param sim_table: name of file with simulations list and their parameters
    :param data_table: name of file with simulation results and processed metrics
    :param cols_to_negate: if not None, then negate specified columns
    :param abs_cols: if not None, then replace values in specified columns with their absolute values
    """
    add_data = read_sim_table(os.path.join(root_folder, sim_table)).set_index("N")
    main_data = get_sim_data(
        os.path.join(root_folder, data_table),
        cols_to_negate=cols_to_negate,
        abs_cols=abs_cols,
    ).set_index("exp_num")

    return main_data.join(add_data, how="left").reset_index()


def full_simulation_loading(
    root_folder: str, files_dict: Dict[str, str], return_ndf: bool = True
) -> pd.DataFrame:
    """Loads multiple batches of results into narrow DataFrame with preprocessing.

    NOTE: in CSV files Ftot = set force clamp, DOUBLE CHECK

    :param root_folder: folder where all sims are stored
    :param files_dict: dict {<sim summary XLSX>: <results CSV>}
    :param return_ndf: if True, returns narrow DataFrame, otherwise returns wide DataFrame
    :return: a DataFrame with results in narrow format
    """
    df_lst = [
        get_joined_data(
            root_folder=root_folder,
            sim_table=sfd,
            data_table=dfd,
            cols_to_negate=default_cols_to_negate,
            abs_cols=default_abs_cols,
        )
        for sfd, dfd in files_dict.items()
    ]

    df = pd.concat(df_lst, axis=0, ignore_index=True, sort=False).rename(
        columns={"Ftot": "F"}
    )

    print(get_contigious_ranges(df["exp_num"].unique()))

    if return_ndf:
        ndf = to_narrow(df).dropna(axis=0).sort_values("exp_num")
        ndf["avg_tot"] = ndf["avg_dG"] + ndf["G"]
        ndf = add_asymmetry_narrow(ndf, "exp_num")
        return ndf
    return df


def read_source(path: str) -> pd.DataFrame:
    """Read input file from `path` according to its extension.

    Supported extensions:
    - 'xls' or 'xlsx'
    - 'csv'

    :param path: file to read
    :return: read data as pd.DataFrame
    """
    _path_check = path.lower()
    if _path_check.endswith(".xls") or _path_check.endswith(".xlsx"):
        return pd.read_excel(path)
    elif _path_check.endswith(".csv"):
        return pd.read_csv(path)
    else:
        raise TypeError("Currently supports only XLS, XLXS or CSV input formats")


def get_exp_data(path: str, narrow: bool = True) -> pd.DataFrame:
    """Get experimental data (VD)

    :param path: file with experimental data to download
    :param narrow: if True, convert to narrow format (labels for direction, positive F and v)
    :return: experimental data as pd.DataFrame
    """
    df = read_source(path)

    split_num = df["exp_name"].str.rsplit("-", expand=True)
    df["pedestal"] = split_num.iloc[:, 0].str.cat(split_num.iloc[:, 1], sep="-")
    split_num.iloc[:, 1] = split_num.iloc[:, 1].str[:2]
    df["dumbbell"] = split_num.iloc[:, 0].str.cat(split_num.iloc[:, 1], sep="-")

    if narrow:
        df = to_narrow(df)
    return df


def full_experimental_loading(
    main_exp_data_path: str, corr_forces_path: str
) -> pd.DataFrame:
    """Download experimental data and preprocess it.

    :param main_exp_data_path: path to file with main experimental results
    :param corr_forces_path: path to file with correct forces (calculated with UFFC mechanical model)
    :return: prepared experimental data
    """
    exp_data = get_exp_data(main_exp_data_path)

    corr_forces = get_exp_data(corr_forces_path)
    exp_data = exp_data.merge(
        corr_forces, how="left", on=["exp_name", "pedestal", "dumbbell", "direction"]
    )
    exp_data.rename(
        columns={
            "Fmol": "Fmol_bad",
            "Fmol_corr": "Fmol",
            "Ftot": "Ftot_bad",
            "Ftot_corr": "Ftot",
        },
        inplace=True,
    )

    strange_exp_num = exp_data[exp_data["Fmol"] > exp_data["Ftot"]]["pedestal"].unique()
    print(
        "Fraction of experiments with Fmol > Ftot",
        exp_data["pedestal"].isin(strange_exp_num).mean(),
    )
    exp_data = exp_data[~exp_data["pedestal"].isin(strange_exp_num)]

    exp_data["chamber"] = exp_data["dumbbell"].str[:-1]
    exp_data = add_asymmetry_narrow(exp_data, "exp_name")
    prepare_nonlinear_cols(exp_data)

    return exp_data


def get_rates_info(path: str) -> pd.DataFrame:
    """Get fitted rate of on / off dynamics of the second domain.

    :param path: path to file with rates
    """
    rename_map = {"side": "direction"}
    df = pd.read_csv(path, na_values=["None"])
    return df.rename(columns=rename_map)
