"""Module with plotting functions."""

import re
from typing import Any, Dict, List, Union, Tuple, Optional, Collection

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from matplotlib.axes import Axes
from matplotlib.figure import Figure
from scipy.interpolate import griddata, PchipInterpolator
from statsmodels.nonparametric.smoothers_lowess import lowess

from uffc.constants import kT
from uffc.utils import check_argument


def fin_plot() -> None:
    """Shows and closes current plot."""
    plt.show()
    plt.close()


def create_axes_if_none(ax: Optional[Axes], **options) -> Tuple[Figure, Axes]:
    """Creates new Figure-Axes pair if passed ax=None, otherwise return existing."""
    if ax is None:
        return plt.subplots(**options)
    return ax.get_figure(), ax


def legend_subtitle(ax: Axes, title: str, kind: str = "scatter") -> None:
    """Adds subtitle to the legend.

    This function must be called right before the plot being annotated.

    :param ax: Axes where dat and legend should be placed
    :param title: subtitle to insert in legend
    :param kind: kind of plots this header precedes. Must be one of 'scatter' or 'line'
    """
    if kind == "scatter":
        ax.scatter([], [], s=0, label=title)
    elif kind == "line":
        ax.plot([], [], lw=0, label=title)
    else:
        raise ValueError(f'Unknown kind {kind}, should be one of "scatter" or "line".')


def plot_bisector(ax: Axes, **kwargs) -> Axes:
    """Plots bisector on a given Axes.

    By default plots red 0.5pt dashed line. Other settings may be provided as kwargs.
    Plotting range: (min(xmin, ymin), max(xmax, ymax))

    :param ax: Axes to draw bisector in
    :param kwargs: kwargs for plt.plot function to tweak bisector appearance.
    :return: modified Axes
    """
    xlim, ylim = ax.get_xlim(), ax.get_ylim()
    lim = [min(xlim[0], ylim[0]), max(xlim[1], ylim[1])]

    params = dict(c="r", ls="--", lw=0.5)
    params.update(kwargs)

    ax.plot(lim, lim, **params)
    return ax


def place_legend(
    ax: Axes, fmt: str = None, loc="center left", bbox_to_anchor=(1.05, 0.5), **kwargs
) -> Axes:
    """Places and formats the legend.

    By default puts it centered to the right from the Axes.

    :param ax: Axes for which to change legend
    :param fmt: optional format string that is allied to numeric values. If None, no formatting is performed
    :param loc: location of the legend
    :param bbox_to_anchor: BBox to anchor the legend
    :param kwargs: additional keyword arguments passed to legend constructor
    """
    handles, labels = ax.get_legend_handles_labels()
    if (lgnd := ax.get_legend()) is not None:
        title = lgnd.get_title().get_text()
    else:
        title = None

    if fmt is not None:
        fmtd_lbls = []
        for l in labels:
            try:
                fmtd_lbls.append(f"{float(l):{fmt}}")
            except ValueError:
                fmtd_lbls.append(l)
    else:
        fmtd_lbls = labels
    ax.legend(
        handles,
        fmtd_lbls,
        title=title,
        loc=loc,
        bbox_to_anchor=bbox_to_anchor,
        **kwargs,
    )
    return ax


def plot_axes(
    ax: Axes,
    plot_x: bool = True,
    plot_y: bool = True,
    options: Optional[Dict[str, Any]] = None,
) -> Axes:
    """Draws X or Y axes of the plot."""
    if options is None:
        options = dict(c="k", lw=0.5)
    if plot_x:
        ax.axhline(0, **options)
    if plot_y:
        ax.axvline(0, **options)
    return ax


def plot_segments(
    df: pd.DataFrame,
    x: str,
    upper: str,
    lower: str,
    colors: str,
    ax: Optional[Axes] = None,
    **kwargs,
) -> Axes:
    """Plots vertical segments.

    :param df: data to plot
    :param x: x values column name
    :param upper: name of column with upper segemnt ends
    :param lower: name of column with lower segemnt ends
    :param colors: color column name
    :param ax: axes to draw, defaults to None (creates default figure-axes pair)
    :return: modified axes with segemnt plot
    """
    fig, ax = create_axes_if_none(ax)

    df.plot.scatter(x=x, y=lower, ax=ax, **kwargs)
    df.plot.scatter(x=x, y=upper, ax=ax, **kwargs)

    ax.vlines(df[x], df[lower], df[upper], color=colors)

    return ax


def assure_zero_origin(ax: Axes) -> None:
    """Adjusts Axes limits so that (0, 0) point is visible."""
    val_min, val_max = ax.get_xlim()
    if val_max < 0:
        ax.set_xlim((None, 0))
    elif val_min > 0:
        ax.set_xlim((0, None))

    val_min, val_max = ax.get_ylim()
    if val_max < 0:
        ax.set_ylim((None, 0))
    elif val_min > 0:
        ax.set_ylim((0, None))


def _get_plot_range(arr: Union[pd.Series, np.ndarray]) -> Tuple[float, float]:
    """For a given array calculates range based on integer grid

    :param arr: input array-like data
    :return: tuple min, max
    """
    return np.floor(arr.min()), np.ceil(arr.max())


def _get_plot_range_exp(ser: Union[pd.Series, np.ndarray]) -> Tuple[float, float]:
    """For a given array calculates range based on integer grid for values mantissa

    :param ser: input array-like data
    :return: tuple min, max
    """
    power = np.floor(np.log10(ser.max()))
    ex = 10**power
    mant = ser / ex
    return np.floor(mant.min()) * ex, np.ceil(mant.max()) * ex


def plot_rates(ndf: pd.DataFrame, average: bool = False):
    minG, maxG = _get_plot_range(ndf["rate"])

    if average:
        df = (
            ndf.groupby(["direction", "kind", "B", "G_rot"])["rate"]
            .mean()
            .reset_index()
        )
    else:
        df = ndf

    gr = sns.FacetGrid(
        df,
        row="direction",
        col="kind",
        hue="G_rot",
        palette="viridis",
        height=5,
        ylim=(minG, maxG),
    )

    def _plotter(x, y, **kws):
        sns.swarmplot(y=y, x=x, **kws)

    gr.map(_plotter, "B", "rate").add_legend()
    plt.yscale("log")


def plot_rates_validation(ndf: pd.DataFrame) -> None:
    gr = ndf.groupby(["direction", "kind"])

    for name, g in gr:
        sns.scatterplot(
            data=g,
            x="n",
            y="rate",
            hue="G_rot",
            size="G",
            palette="winter",
            legend="full",
        )

        #         plt.legend(vmax=8, vmin=6)
        plt.yscale("log")
        plt.legend(bbox_to_anchor=(1.05, 0.5), loc="center left")
        plt.title(name)

        fin_plot()


def plot_n_validation(ndf: pd.DataFrame) -> None:
    gr = ndf.groupby(["direction", "kind"])

    for name, g in gr:
        sns.scatterplot(
            data=g,
            y="n",
            x="v",
            size="G_rot",
            hue="B",
            palette="viridis",
            legend="full",
        )
        #         plt.legend(vmax=8, vmin=6)
        #         plt.yscale('log')
        plt.legend(bbox_to_anchor=(1.05, 0.5), loc="center left")
        plt.title("Direction {}, k_{}".format(*name))

        fin_plot()


def plot_rss(ndf: pd.DataFrame) -> None:
    minG, maxG = _get_plot_range_exp(ndf["RSSn"])

    gr = sns.FacetGrid(
        ndf[ndf["kind"] == "on"],
        row="direction",
        col="G_rot",
        hue="G",
        palette="inferno",
        height=5,
        ylim=(minG, maxG),
    )

    def _plotter(x, y, **kws):
        sns.swarmplot(x=x, y=y, **kws)

    gr.map(_plotter, "B", "RSSn", alpha=0.75).add_legend()


def plot_avg_tot_G_LR(
    df: pd.DataFrame, hue_col: str = "Ftot", size_col: Optional[str] = None
):
    """Plots dependency of apparent average well depth in R and L directions.

    :param hue_col: feature to use for color coding
    :param size_col: feature to user for size coding
    """

    minG, maxG = _get_plot_range(df["avg_dG"])

    for g in df["G"].unique():
        tdf = df[df["G"] == g].drop_duplicates(subset=["exp_num", "direction"])
        tdf = tdf.pivot(index="exp_num", columns="direction", values="avg_dG").merge(
            tdf[["exp_num", hue_col, size_col]].drop_duplicates(), on="exp_num"
        )

        fig, ax = plt.subplots(figsize=(6, 6))
        sns.scatterplot(
            data=tdf, x="R", y="L", hue=hue_col, size=size_col, ax=ax, palette="cool"
        )

        ax.plot([minG, maxG], [minG, maxG], c="lightgrey", linestyle="-.")
        ax.set_xlim(minG, maxG)
        ax.set_ylim(minG, maxG)
        plt.title("Depth of the main well {} kT".format(g))
        plt.legend(bbox_to_anchor=(1.05, 0.5), loc="center left")

        plt.xlabel("Average depth during motion to the Right, kT")
        plt.ylabel("Average depth during motion to the Left, kT")

        plt.show()
        plt.close()


def plot_force_vel_G_group(ndf: pd.DataFrame) -> None:
    gb_key = ["G"]
    gg = ndf.groupby(gb_key)

    minG, maxG = _get_plot_range(ndf["avg_tot"])

    for name, g in gg:
        fig, ax = plt.subplots(figsize=(10, 6))

        g["B"] = g["B"].astype(str)

        sns.scatterplot(
            data=g[g["direction"] == "R"],
            x="Fmol",
            y="v",
            style="B",
            hue="avg_tot",
            ax=ax,
            palette="Reds",
            hue_norm=(minG, maxG),
        )

        sns.scatterplot(
            data=g[g["direction"] == "L"],
            x="Fmol",
            y="v",
            style="B",
            hue="avg_tot",
            ax=ax,
            palette="Blues",
            hue_norm=(minG, maxG),
        )

        ax.set_xlim(0, 8)
        ax.set_ylim(0, 200)
        ax.legend(bbox_to_anchor=(1.05, 0.5), loc="center left")
        ax.set_title("Main well depth {} kT".format(name))

        ax.set_xlabel("Force on the molecule, pN")
        ax.set_ylabel("Velocity, µm/s")

        fin_plot()


def plot_mean_avg_dG(ndf: pd.DataFrame) -> None:
    avg_depth = (
        ndf.groupby(["k_rot", "G_rot", "Ftot", "direction"])["avg_dG"]
        .agg(["mean", "std", "count"])
        .reset_index()
    )

    Gmin, Gmax = _get_plot_range(ndf["G_rot"])

    gg = avg_depth.groupby(["direction"])

    for name, g in gg:
        grid = sns.lmplot(
            data=g, x="G_rot", y="mean", hue="Ftot", legend=True, fit_reg=False
        )

        grid.ax.plot([Gmin, Gmax], [Gmin, Gmax], color="grey")

        plt.title("Direction {}".format(*name))
        plt.xlim(Gmin, Gmax)
        #     plt.ylim(0, Gmax)

        fin_plot()


def plot_avg_depth_barplot(ndf: pd.DataFrame) -> None:
    avg_depth = (
        ndf.groupby(["k_rot", "G_rot", "Ftot", "direction"])["avg_dG"]
        .agg(["mean", "std", "count"])
        .reset_index()
    )
    for side in ["R", "L"]:
        tdf = avg_depth[avg_depth["direction"] == side]
        colors = list(mcolors.BASE_COLORS.keys())

        fig, ax = plt.subplots(figsize=(8, 6))
        gr = tdf.groupby("Ftot")
        for i, gg in enumerate(gr):
            force, g = gg
            g["sem"] = g["std"] / np.sqrt(g["count"])
            g.plot(
                x="G_rot",
                y="mean",
                yerr="sem",
                kind="bar",
                width=0.1,
                ax=ax,
                label="{} pN".format(force),
                position=i,
                color=colors[i],
            )

        ax.set_title("Direction {}".format(side))
        ax.set_xlabel("Set depth of rotational well, kT")
        ax.set_ylabel("Average apparent rotational well depth, kT")
        fin_plot()


def plot_avg_dG_strip(ndf: pd.DataFrame, hue_col: str = "Ftot") -> None:
    Gmin, Gmax = _get_plot_range(ndf["G_rot"].append(ndf["avg_dG"]))
    for side in ["R", "L"]:
        #         fig, ax = plt.subplots(figsize=(6, 6))
        sdf = ndf[ndf["direction"] == side]
        sns.lmplot(data=sdf, x="G_rot", y="avg_dG", hue=hue_col)

        plt.ylim(Gmin, Gmax)

        plt.title("Direction {}".format(side))
        #         plt.legend(bbox_to_anchor =(1.05, 0.5), loc='center left')
        plt.xlabel("Set depth of rotational well, kT")
        plt.ylabel("Average apparent rotational well depth, kT")

        fin_plot()


def plot_force_vel_ndf(
    df: pd.DataFrame,
    hue_col: str,
    size_col: Optional[str] = None,
    opacity: float = 0.5,
    discrete_hue: bool = True,
    contour: bool = False,
    kind: str = "linear",
    unite: bool = False,
) -> None:
    if discrete_hue:
        pal_r = sns.color_palette("Set1", df[hue_col].nunique())
        pal_l = sns.color_palette("Dark2", df[hue_col].nunique())
    else:
        pal_r = "Reds"
        pal_l = "Blues"

    df_R = df[df["direction"] == "R"]
    df_L = df[df["direction"] == "L"]

    fig, ax = plt.subplots(figsize=(10, 8))

    if unite:
        lst = []
        if size_col is not None:
            lst += [size_col]
        if hue_col is not None:
            lst += [hue_col]
        cols_to_select = ["Fmol", "v"] + lst
        tmp_df = df[cols_to_select]
        sns.scatterplot(
            data=tmp_df,
            x="Fmol",
            y="v",
            hue=hue_col,
            size=size_col,
            palette=pal_r,
            alpha=opacity,
            ax=ax,
        )
    else:
        sns.scatterplot(
            data=df_R,
            x="Fmol",
            y="v",
            hue=hue_col,
            size=size_col,
            palette=pal_r,
            alpha=opacity,
            ax=ax,
        )
        sns.scatterplot(
            data=df_L,
            x="Fmol",
            y="v",
            hue=hue_col,
            size=size_col,
            ax=ax,
            palette=pal_l,
            alpha=opacity,
        )

    if contour:
        grid_x, grid_y = np.mgrid[0 : df["Fmol"].max() : 20j, 0 : df["v"].max() : 20j]

        zi = griddata(
            df_R[["Fmol", "v"]].values,
            df_R["avg_tot"].values,
            (grid_x, grid_y),
            method=kind,
        )
        cnt = ax.contour(grid_x, grid_y, zi, 20, cmap="gist_gray_r")
        plt.colorbar(cnt, ax=ax)

    ax.set_xlim(0, df["Fmol"].max())
    ax.set_ylim(0, df["v"].max())
    ax.legend(bbox_to_anchor=(1.5, 0.5), loc="center left")
    ax.set_xlabel("Force on the molecule, pN")
    ax.set_ylabel("Velocity, µm/s")

    fin_plot()


def plot_families(
    ndf: pd.DataFrame,
    like_res: pd.DataFrame,
    vd_data: pd.DataFrame,
    keys: Optional[Collection[str]],
    palette,
    force_col: str = "Fmol",
    **kwargs,
) -> None:
    """TODO.

    :param ndf:
    :param like_res:
    :param vd_data:
    :param keys:
    :param palette:
    :param force_col:
    :param kwargs:
    """
    if keys is None:
        keys = ["G", "G_rot", "B"]
    df = ndf.merge(like_res, on=keys, how="left")

    fig, ax = plt.subplots()

    sns.scatterplot(data=vd_data, x=force_col, y="v", c=["gray"], ax=ax, **kwargs)
    sns.lineplot(
        data=df,
        x=force_col,
        y="v",
        hue="like",
        palette=palette,
        ax=ax,
        legend=None,
        **kwargs,
    )
    sns.scatterplot(
        data=df, x=force_col, y="v", hue="like", palette=palette, ax=ax, **kwargs
    )

    ax.legend(bbox_to_anchor=(1.05, 0.5), loc="center left")

    ax.set_xlim((0, None))
    ax.set_ylim((0, None))

    fin_plot()


def plot_3d_like(
    data: pd.DataFrame,
    x: str,
    y: str,
    z: str,
    color: str,
    title: str,
    palette,
    highlight_max: bool = False,
    annotate: bool = False,
) -> None:
    """Plots 3D scatter plot with color coded likelihood of the fit

    :param data:
    :param x:
    :param y:
    :param z:
    :param color:
    :param title:
    :param palette:
    :param highlight_max:
    :param annotate:
    """
    fig = plt.figure(figsize=(14, 8))
    ax = fig.add_subplot(111, projection="3d")

    if annotate:
        for vx, vy, vz, vc in data[[x, y, z, color]].itertuples(index=False):
            ax.text(vx, vy, vz, "{:.3g}".format(vc), size=10, zorder=1, color="k")

    if highlight_max:
        max_idx = data[color].idxmax()
        max_row = tuple(data[[x, y, z]].loc[max_idx, :])
        ax.scatter(*max_row, c="r", depthshade=False)

        data = data[data.index != max_idx]

    scatter = ax.scatter(
        data[x], data[y], data[z], c=data[color], depthshade=False, cmap=palette
    )

    ax.set_xlabel(x)
    ax.set_ylabel(y)
    ax.set_zlabel(z)

    cbar = fig.colorbar(scatter)
    cbar.set_label(color, rotation=270)
    cbar.ax.get_yaxis().labelpad = 15

    ax.set_title(title)

    #     plt.legend(bbox_to_anchor =(1.05, 0.5), loc='center left')
    fin_plot()


def fancy_plot(
    func,
    *args,
    figsize=(20, 10),
    xlim=(0, None),
    ylim=(0, None),
    xscale="linear",
    yscale="linear",
    return_axes=False,
    ax=None,
    **kwargs,
):
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    func(*args, **kwargs, ax=ax)

    ax.set_xscale(xscale)
    ax.set_yscale(yscale)

    if xscale == "log":
        if xlim[0] <= 0:
            xlim = (None, xlim[1])

    if yscale == "log":
        if ylim[0] <= 0:
            ylim = (None, ylim[1])

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    ax.legend(bbox_to_anchor=(1.05, 0.5), loc="center left")

    if return_axes:
        return ax
    else:
        fin_plot()


def generate_splines(
    df: pd.DataFrame,
    smoothing: float = 0.4,
    xcol: str = "Fmol",
    ycol: str = "v",
    include_origin: bool = True,
    independent_var: str = "x",
) -> PchipInterpolator:
    """Draws a spline through LOWESS-smoothed data points.

    :param df: input DataFrame
    :param smoothing: `frac` parameter for LOWESS smoothing
    :param xcol: name of X values column
    :param ycol: name of Y values column
    :param include_origin: whether to add (0, 0) point to data prior to smoothing
    :param independent_var: independent variable for LOWESS; defaults to 'x'
    :return: PchipInterpolator spline object
    """
    if independent_var not in set("xy"):
        raise ValueError("Parameter `independent_var` must be one of 'x' or 'y'")
    if independent_var == "x":
        dep_col, indep_col = ycol, xcol
    else:
        dep_col, indep_col = xcol, ycol

    data = df[[xcol, ycol]]
    if include_origin:
        data = data.append({xcol: 0, ycol: 0}, ignore_index=True)
    data = data.sort_values(indep_col)

    sm_data = lowess(
        data[dep_col].to_numpy(), data[indep_col].to_numpy(), frac=smoothing
    )
    spl = PchipInterpolator(sm_data[:, 0], sm_data[:, 1])

    return spl


def connect_with_splines(
    df: pd.DataFrame,
    smoothing: float = 0.4,
    xcol: str = "Fmol",
    ycol: str = "v",
    include_origin: bool = True,
    independent_var: str = "x",
    ax: Optional[Axes] = None,
    label: str = "",
    **kwargs,
) -> Axes:
    """Draws a spline through LOWESS-smoothed data points.

    :param df: input DataFrame
    :param smoothing: `frac` parameter for LOWESS smoothing
    :param xcol: name of X values column
    :param ycol: name of Y values column
    :param include_origin: whether to add (0, 0) point to data prior to smoothing
    :param independent_var: independent variable for LOWESS; defaults to 'x'
    :param ax: optional Axes to draw in
    :param label: label for splines to use in legend
    :param kwargs: additional keyword arguments for plt.plot function drawing splines
    :return: Axes with spline drawn
    """
    spl = generate_splines(df, smoothing, xcol, ycol, include_origin, independent_var)
    x = np.linspace(spl.x.min(), spl.x.max(), 50)

    if ax is None:
        fig, ax = plt.subplots()
    if independent_var == "x":
        ax.plot(x, spl(x), **kwargs, label=label)
    else:
        ax.plot(spl(x), x, **kwargs, label=label)
    return ax


def compare_sim_exp(
    ndf: pd.DataFrame,
    vd_data: pd.DataFrame,
    label_cols: List[str],
    sort_by: str = "rank",
    legend: str = "brief",
    alpha: float = 0.3,
    annotate: bool = False,
    figsize: Tuple[float, float] = (10, 10),
    labels: Optional[str] = None,
    diffusion_coeff: Optional[float] = None,
    font_size: int = 14,
    ax: Optional[Axes] = None,
):
    """Force-velocity plot for experimental data and simulations.

    :param ndf: narrow DataFrame with simulations
    :param vd_data: experimental data
    :param label_cols: columns used to label families of simulations in legend. Plotted with different colors.
    :param sort_by: column to sort by simulations
    :param legend: kind of legend to use. See seaborn.scatterplot for details
    :param alpha: opacity for simulation data plot
    :param annotate: whether to annotate each point from simulation data
    :param figsize: size of the resulting figure
    :param labels: optional labels to use for simulation point annotation. Defaults to legend labels
    :param diffusion_coeff: optional diffusion coefficient. If provided, draw low-force diffusive approximation
    :param font_size: font size to use, defaults to 14 pt
    :param ax: optional Axes to draw in. If None, creates a new one with figsize
    """
    fig, ax = create_axes_if_none(ax, figsize=figsize)

    df = ndf.copy()
    df["Fmol"] = df["Fmol"].abs()
    df["v"] = df["v"].abs()

    x = df["Fmol"].values
    y = df["v"].values

    if len(label_cols) > 1:
        sim_id_col = " - ".join(label_cols)

        df[sim_id_col] = df[label_cols[0]].astype(str)
        sep_ser = pd.Series(np.full(len(df), " - "), index=df.index)
        for col in label_cols[1:]:
            df[sim_id_col] += sep_ser + df[col].astype(str)
    else:
        sim_id_col = label_cols[0]

    df.sort_values(sort_by, inplace=True)

    legend_subtitle(ax, "Experimental data")
    sns.scatterplot(
        data=vd_data,
        x="Fmol",
        y="v",
        hue="direction",
        style="direction",
        markers=["P", "X"],
        palette="RdBu",  # 'bwr', # 'spring',
        alpha=1,
        legend=legend,
        ax=ax,
    )

    #     ax.errorbar(x=vd_data['Fmol'], y=vd_data['v'], xerr=0.5, yerr=10, marker='', ls='', lw=0.2, c='k', alpha=1)

    legend_subtitle(ax, "Simulations")
    #     legend_subtitle(ax, sim_id_col)

    #     for dd, gg in df.groupby('direction'):
    #         sns.scatterplot(data=gg, x='Fmol', y='v',
    #                         hue=sim_id_col,
    #                         palette='tab10',
    #                         alpha=alpha,
    #                         legend=(legend if dd == 'L' else None),
    #                         estimator=None,
    #                         s=120,
    #                         marker=('^' if dd == 'L' else '.'),
    #                         ax=ax,
    #                     )

    _markers = ["o", "^"] if df["direction"].nunique() > 1 else ["o"]

    sns.scatterplot(
        data=df,
        x="Fmol",
        y="v",
        hue=sim_id_col,
        palette="tab10",
        alpha=alpha,
        # legend=(legend if dd == 'L' else None),
        legend=legend,
        estimator=None,
        s=60,
        style="direction",
        markers=_markers,
        ax=ax,
    )

    if diffusion_coeff is not None:
        max_force = df["Fmol"].max()
        ax.plot(
            [0, max_force],
            [0, max_force / (kT / diffusion_coeff)],
            c="r",
            ls="--",
            lw=0.75,
        )

    if annotate:
        if labels is None:
            labels = df[sim_id_col]
        for i, txt in enumerate(labels):
            ax.annotate(txt, (x[i], y[i]), xytext=(5, 5), textcoords="offset points")

    ax.legend(bbox_to_anchor=(1.05, 0.5), loc="center left")
    ax.set_xlabel("Force on the molecule, pN", fontsize=font_size)
    ax.set_ylabel("Dumbbell velocity, µm/s", fontsize=font_size)
    ax.tick_params(labelsize=font_size)

    ax.set_title(
        "Comparison between experimental data and simulations", fontsize=font_size
    )
    ax.set_xlim((0, None))
    ax.set_ylim((0, None))

    return ax


def create_figure(
    fig_width: Union[float, str] = 2.25,
    aspect: float = 4.0 / 3,
    unit: str = "inches",
    nrows: Optional[int] = None,
    ncols: Optional[int] = None,
    **kwargs,
) -> Tuple[Figure, Union[Axes, np.ndarray]]:
    """Creates figure with given width and aspect ratio.

    :param fig_width: value of figure width. If float, use `unit` option, if str, override `unit` option.
    :param aspect: aspect ratio (width : height)
    :param unit: unit of the width. Must be one of 'inches' or 'cm'. Used only with float `fig_width`
    :param nrows: optional number of rows
    :param ncols: optional number of columns
    :param kwargs: additional kwargs for `plt.subplots`
    :return: tuple of figure and corresponding axes
    """
    if isinstance(fig_width, str):
        pattern = re.compile("^(?P<width>[\d.]+)(?P<unit>[a-z]+)$")
        parsed = pattern.match(fig_width).groupdict()
        fig_width = float(parsed["width"])
        unit = parsed["unit"]

    check_argument(unit, "unit", {"cm", "inches", "in"})
    if isinstance(fig_width, (float, int)):
        if unit == "cm":
            fig_width /= 2.54

    fig_height = fig_width / aspect
    fig_size = (fig_width, fig_height)

    if (nrows is None) ^ (ncols is None):
        raise ValueError("ncols and nrows must be None or not None simultaneously.")
    elif nrows is None:
        return plt.subplots(figsize=fig_size, **kwargs)
    return plt.subplots(nrows, ncols, figsize=fig_size, **kwargs)


def scale_text_size(factor: float) -> None:
    """Scales text size from current values, round upwards."""
    for prop in [
        "axes.titlesize",
        "axes.labelsize",
        "xtick.labelsize",
        "ytick.labelsize",
        "legend.fontsize",
        "legend.title_fontsize",
    ]:
        plt.rcParams[prop] = np.ceil(plt.rcParams[prop] * factor)


def set_text_size(size: int) -> None:
    """Sets text size."""
    for prop in [
        "axes.titlesize",
        "axes.labelsize",
        "xtick.labelsize",
        "ytick.labelsize",
        "legend.fontsize",
        "legend.title_fontsize",
    ]:
        plt.rcParams[prop] = size
