"""Module with tools for simulation results loading and analysis."""

import os
from dataclasses import asdict, dataclass
from typing import Callable, Collection, Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import simplejson
from joblib import Parallel, delayed
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from tqdm import tqdm

from uffc.constants import kT
from uffc.plotting import fin_plot
from uffc.utils import check_argument


def get_default_design_dict(
    n_mols: int,
) -> Dict[str, Tuple[str, str, Optional[Tuple], Optional[Callable]]]:
    """Factory that generates default dict with settings for feature plotting.

    :param n_mols: number of molecule in simulation
    :return: dict `feature_name`: tuple with settings:
    - title of the plot
    - Y-axis label
    - Y-axis limits
    - Y-axis transformation (such as um -> nm)
    """

    def scaler_1k(x):
        return x * 1000

    pos_label = "Position (nm)"

    design_dict = {
        "xBeadr": ("Position of the right bead", pos_label, None, scaler_1k),
        "xBeadl": ("Position of the left bead", pos_label, None, scaler_1k),
        "xTrapr": ("Position of the right trap", pos_label, None, scaler_1k),
        "xTrapl": ("Position of the left trap", pos_label, None, scaler_1k),
        "xMT": ("Position of the MT center", pos_label, None, scaler_1k),
    }
    if n_mols == 1:
        design_dict.update(
            {
                "xMol": (
                    "Position of the molecule relative to the pedestal",
                    pos_label,
                    None,
                    scaler_1k,
                ),
                "phi": (
                    "Angle between the molecule and the MT",
                    "Angle (°)",
                    None,
                    lambda x: x * 180 / np.pi,
                ),
                "deltaG": (
                    "Energy of 2nd site interaction with the MT",
                    "Energy ($k_B T$)",
                    None,
                    lambda x: x / kT,
                ),
                "logpotentialForce": (
                    "Force of the molecule-MT interaction",
                    "Force (pN)",
                    None,
                    None,
                ),
                "binding": (
                    "Binding state (1 - bound, 0 - unbound)",
                    "Binding state",
                    None,
                    None,
                ),
            }
        )
    else:
        for n in range(1, n_mols + 1):
            design_dict.update(
                {
                    f"xMol{n}": (
                        f"Position of the molecule #{n} relative to the pedestal",
                        pos_label,
                        None,
                        scaler_1k,
                    ),
                    f"phi{n}": (
                        f"Angle between the molecule #{n} and the MT",
                        "Angle (°)",
                        None,
                        lambda x: x * 180 / np.pi,
                    ),
                    f"deltaG{n}": (
                        f"Energy of interaction between 2nd site of molecule #{n} and the MT",
                        "Energy ($k_B T$)",
                        None,
                        lambda x: x / kT,
                    ),
                    f"logpotentialForce{n}": (
                        f"Force of the molecule #{n} - MT interaction",
                        "Force (pN)",
                        None,
                        None,
                    ),
                    f"binding{n}": (
                        f"Binding state (1 - bound, 0 - unbound) of molecule #{n}",
                        "Binding state",
                        None,
                        None,
                    ),
                }
            )
    return design_dict


class ResultsStorage:
    """Class to access results stored on a disk.

    Example of the tree structure:

    results-3419-3434                          # folder with a run
    |                                          # name structure: results-<start_exp_num>-<end_exp_num>
    |- 3419_desktop-9qihk4d                    # folder with results of a single simulation
        |                                      # name structure: <exp_num>_<pc_name>
        |- 2s_binOff_illustrative_single.json  # task file
        |- 3419.json                           # config file
        |- 3419_results_xBeadr.binary          # files with data
        |- ...
        |- Langevin2.exe                       # executable for the simulator
    |- 3420_desktop-9qihk4d
    |- ...
    """

    def __init__(self):
        self.data = dict()

    @classmethod
    def from_tuples(cls, runs: Collection[Tuple[str, str]]) -> "ResultsStorage":
        """Creates a new instance from folder and PC name tuples."""
        self = cls()
        for path, pc_name in tqdm(runs):
            if path[-1] not in {"/", "\\"}:
                raise ValueError("Path to folders must end with slash!")
            rng = tuple(
                map(int, os.path.basename(os.path.dirname(path)).split("-")[1:])
            )
            self.data[rng] = (path, pc_name)
        return self

    def __contains__(self, item: int) -> bool:
        """Checks whether a simulation with a given ID is in the storage."""
        for start, end in self.data.keys():
            if start <= item <= end:
                return True
        return False

    def __iter__(self) -> Tuple[int, Tuple[str, str]]:
        for start, end in self.data.keys():
            obj = self.data[(start, end)]
            for i in range(start, end + 1):
                yield i, obj


@dataclass
class SimulationParameters:
    """Dataclass for simulation parameters."""

    exp_num: int

    F: float
    Fpre: float  # prestretch force parameter, equals 2 * real prestretch force
    amplitude: float
    k1: float  # stiffness of the left trap
    k2: float  # stiffness of the right trap

    G: float
    G_rot: float
    k_rot: float
    B: float  # relative barrier height; in simulations for the paper B = 0

    A: float  # width parameter of an asymmetric well
    m: float  # asymmetry parameter of an asymmetric well between -1 and 0

    gammaBeadL: float
    gammaBeadR: float
    gammaMT: float
    gamma_dmbl: float

    # ========== THESE ATTRIBUTES ARE FOR TWO MOLECULE SIMULATIONS ==========
    # They are not used in simulations with a single molecule or
    # if a bindingDynamics is set to 0 (disabled)
    # These attributes are set to None in single molecule simulations
    kOn1: Optional[float] = None  # attachment rate for the first molecule
    kOff1: Optional[float] = None  # detachment rate for the first molecule
    kOn2: Optional[float] = None  # attachment rate for the second molecule
    kOff2: Optional[float] = None  # detachment rate for the second molecule
    # =======================================================================

    gitHashShort: Optional[str] = None
    # optional hash of the Git commit corresponding to the simulation code

    def as_dict(self) -> Dict:
        """Converts dataclass into dict."""
        return asdict(self)

    @classmethod
    def from_config(cls, config: Dict):
        """Creates SimulationParameters class from simulation config dict.

        Parameter ``Fpre`` is converted into real prestretch force, i.e. the Fpre / 2
        """
        poi = {
            "Gdepth": "G",
            "movementTotalForce": "F",
            "prestretchTotalForce": "Fpre",
            "rotWellDepth": "G_rot",
            "rotStiffness": "k_rot",
            "trapstiffL": "k1",
            "trapstiffR": "k2",
            "B": "B",
            "gammaBeadL": "gammaBeadL",
            "gammaBeadR": "gammaBeadR",
            "gammaMT": "gammaMT",
            "kOn1": "kOn1",
            "kOff1": "kOff1",
            "kOn2": "kOn2",
            "kOff2": "kOff2",
            "DmblMoveAmplitude": "amplitude",
            "A": "A",
            "m": "m",
        }

        params = config["ModelParameters"]
        res = {poi[k]: float(v) for k, v in params.items() if k in poi}

        res["gitHashShort"] = config.get("gitHashShort")

        res["exp_num"] = int(config["Name"])
        res["Fpre"] = res["Fpre"] / 2
        res["gamma_dmbl"] = res["gammaBeadL"] + res["gammaBeadR"] + res["gammaMT"]
        res["F"] = abs(res["F"])
        return cls(**res)


class Simulation:
    """Class encapsulations simulation results."""

    def __init__(self, folder: str, pc_name: str, exp_num: int) -> None:
        """Initializes a new instance of the Simulation object.

        :param folder: folder with results of the simulation
        :param pc_name: name of the machine where computations took place (used as part of file names)
        :param exp_num: simulation ID (number, label) to read
        """
        self.folder = folder
        self.pc_name = pc_name
        self.exp_num = exp_num

        self.config = dict()
        self.data = dict()

    @classmethod
    def load_simulation(
        cls, folder: str, pc_name: str, exp_num: int, feature_list: Collection[str]
    ) -> "Simulation":
        """Loads simulation data from folder with binary files and configs.

        :param folder: folder where all simulation files are stored
        :param pc_name: name of the machine where computations took place (used as part of file names)
        :param exp_num: simulation ID (number, label) to read
        :param feature_list: list of features to download
        :return: new Simulation object
        """
        self = cls(folder, pc_name, exp_num)
        self.load_config()
        self.data = {feature: self.load_feature(feature) for feature in feature_list}
        return self

    def load_config(self) -> Dict[str, Union[Dict, str]]:
        """Load config into self and return contents.

        :return: configuration of the simulation
        """
        cfg_path = os.path.join(
            self.folder,
            "{}_{}".format(self.exp_num, self.pc_name),
            "{}.json".format(self.exp_num),
        )
        with open(cfg_path) as fd:
            config = simplejson.load(fd)
        self.config = config["Configuration"]
        return self.config

    def load_feature(self, feature: str) -> np.ndarray:
        """Loads specified feature from binary results file.

        :param feature: feature name to read from file
        :return: np.ndarray with data
        """
        dct = dict(exp_num=self.exp_num, pc=self.pc_name, feature=feature)
        file_path = os.path.join(
            self.folder,
            "{exp_num}_{pc}".format(**dct),
            "{exp_num}_results_{feature}.binary".format(**dct),
        )
        if not os.path.exists(file_path):
            print(f"File {file_path} does not exist, skipping")
        return np.fromfile(file_path)

    def shift_features(
        self, feature_list: Collection[str], inplace: bool = False
    ) -> Optional["Simulation"]:
        """Subtract initial value from each feature data array.

        :param feature_list: list of features to alter
        :param inplace: whether modify data inplace or return a copy
        """
        res = {key: self.data[key] - self.data[key][0] for key in feature_list}
        if inplace:
            self.data.update(res)
        else:
            new_sim = Simulation(self.folder, self.pc_name, self.exp_num)
            new_sim.config = self.config.copy()
            new_sim.data = self.data.copy()
            new_sim.data.update(res)
            return new_sim

    def get_parameters(self) -> SimulationParameters:
        """Returns SimulationParameters object from simulation config."""
        return SimulationParameters.from_config(self.config)

    def __getitem__(self, item: str) -> np.ndarray:
        """Gets feature data."""
        return self.data[item]

    def __setitem__(self, key: str, value: np.ndarray) -> None:
        """Sets feature data."""
        self.data[key] = value

    def __delitem__(self, key: str) -> None:
        """Deletes feature."""
        del self.data[key]

    def __contains__(self, item: str) -> bool:
        return item in self.data


class Analysis:
    def __init__(
        self, res_storage: ResultsStorage, feature_list: List[str], kind: str = "uffc"
    ) -> None:
        """

        :param res_storage: object representing results of the simulations
        :param feature_list: list of features (i.e. files with results) to read
        :param kind: must be one of 'all', 'uffc', 'diff'
        """
        self.sims = {}

        def loader(n, folder, pc_name):
            is_bead_file = any(
                "xBeadr" in x
                for x in os.listdir(os.path.join(folder, f"{n}_{pc_name}"))
            )
            if (
                (kind == "all")
                or (kind == "uffc" and is_bead_file)
                or (kind == "diff" and not is_bead_file)
            ):
                self.sims[n] = Simulation.load_simulation(
                    folder=folder, pc_name=pc_name, exp_num=n, feature_list=feature_list
                )
            else:
                print(f"Unknown combination {kind=} and {is_bead_file=}")

        Parallel(n_jobs=6, require="sharedmem")(
            delayed(loader)(n, folder, pc_name)
            for n, (folder, pc_name) in tqdm(res_storage)
        )

    def get_parameters_table(self) -> pd.DataFrame:
        """Combine parameters of all simulations of analysis into pd.DataFrame."""
        return pd.DataFrame(
            [sim.get_parameters().as_dict() for sim in self.sims.values()]
        )

    def __getitem__(self, item: int) -> Simulation:
        return self.sims[item]

    def items(self):
        return self.sims.items()

    def values(self):
        return self.sims.values()

    def __iter__(self):
        return iter(self.sims.keys())

    def __contains__(self, item: int) -> bool:
        return item in self.sims

    def timeline_plot(
        self, feature: str, ncols: int = 4, figsize: Tuple[int, int] = (20, 40)
    ) -> None:
        """Plots timelines of the feature for all simulations.

        :param feature: feature to plot
        :param ncols: number of cols to use in the resulting figure
        :param figsize: size of the resulting figure (in inches)
        """
        feature_list = ["Time", feature]

        fig, axs = plt.subplots(
            (len(self.sims) - 1) // ncols + 1, ncols, figsize=figsize
        )
        for ax, (i, sim_raw) in zip(axs.ravel(), enumerate(self.values())):
            sim = sim_raw.shift_features(feature_list, inplace=False)

            ax.plot(sim["Time"], sim[feature])
            ax.set_title(sim.exp_num)
            ax.set_xlim((0, None))
        fin_plot()


# TODO refactor as Simulation method?
def plot_raw_data(
    fin_sim: Simulation,
    feature_list: List,
    ncols: int = 3,
    design_dict: Optional[Dict[str, Tuple]] = None,
    suptitle: Optional[str] = None,
    cell_size: Tuple[float, float] = (8, 6),
    kind: str = "scatter",
    time_axis_label: str = "Time, ms",
) -> Tuple[Figure, List[Axes]]:
    """Plots timelines of several features of a single simulation.

    :param fin_sim: Simulation object to use
    :param feature_list: list of features to plot timelines for
    :param ncols: number of columns in the resuling figure
    :param design_dict: dictionary with design settings. See ``get_default_design_dict``
        function for details.
    :param suptitle: figure suptitle to use
    :param cell_size: size of each plot cell (in inches)
    :param kind: type of the plot to use, defaults to 'scatter'. Must be one of
        'scatter', 'line'.
    :param time_axis_label: label to use for the time axis
    :return: a 2-tuple of the Figure object and Axes array
    """
    check_argument(kind, "kind", {"scatter", "line"})

    if design_dict is None:
        design_dict = get_default_design_dict(1)

    features_to_plot = sorted(set(feature_list).difference({"Time", "direction"}))

    nrows = (len(features_to_plot) - 1) // ncols + 1
    figsize = (cell_size[0] * ncols, cell_size[1] * nrows)
    fig, axs = plt.subplots(ncols=ncols, nrows=nrows, figsize=figsize)
    if suptitle is None:
        fig.suptitle(f"Simulation {fin_sim.exp_num}")

    rel_offset = 1.2  # for direction pulse

    time_arr = fin_sim["Time"] * 1000
    for ax, feat in zip(axs.ravel(), features_to_plot):
        title, ylabel, ylim, ytransform = design_dict[feat]

        arr = fin_sim[feat]
        if ytransform:
            arr = np.apply_along_axis(ytransform, axis=-1, arr=arr)

        ymin, ymax = arr.min(), arr.max()
        indicator = (
            fin_sim["direction"] * (ymax - ymin) * rel_offset / 2 + (ymax + ymin) / 2
        )

        if kind == "line":
            ax.plot(time_arr, arr, lw=0.5)
        elif kind == "scatter":
            ax.scatter(time_arr, arr, s=0.5)
        ax.plot(time_arr, indicator, lw=0.5, c="orange")
        ax.set_title(title)
        ax.set_ylabel(ylabel)

        if ylim:
            ax.set_ylim(ylim)
        #         ax.set_xlim((0, np.ceil(time_arr.max() / 10) * 10)) # TODO refactor
        ax.set_xlim((0, time_arr.max()))
        ax.set_xlabel(time_axis_label)

    if len(features_to_plot) - nrows * ncols > 0:  # there are unused axes
        for ax in axs.ravel()[len(features_to_plot) - nrows * ncols :]:
            ax.set_visible(False)

    plt.subplots_adjust(hspace=0.3)
    return fig, axs
