"""Module with tools for sweeps fitting in simulations and experiments."""

from collections import namedtuple
from copy import deepcopy
from typing import Callable, Dict, List, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from numpy.linalg import LinAlgError
from scipy.linalg import svd
from scipy.optimize import least_squares
from sklearn.metrics import r2_score
from tqdm import tqdm

from uffc.constants import kT
from uffc.get import (
    default_abs_cols,
    default_cols_to_negate,
    set_abs_cols,
    set_negated_cols,
)
from uffc.plotting import fin_plot
from uffc.simulation import Analysis, Simulation, SimulationParameters
from uffc.utils import (
    convert_col_to_types,
    eng_str,
    expand_ranges,
    get_consecutive_pairs,
    swap_dir_prop,
)

Description = namedtuple("Description", ["mean", "sd", "N"])


def describe_array(arr: np.ndarray) -> Description:
    """Returns a named tuple with statistics of the array."""
    dims = arr.ndim
    if dims != 1:
        raise ValueError(f"Input array must be 1D, got {dims}")
    return Description(mean=arr.mean(), sd=arr.std(ddof=1), N=len(arr))


def descriptions_to_record(
    descr_dict: dict[str, Description], suffix: str = ""
) -> dict[str, float]:
    """Flattens a dict with descriptions into a dict with statistics."""
    res = {}
    for feature, desc in descr_dict.items():
        res[feature + suffix] = desc.mean
        res[feature + "_sd" + suffix] = desc.sd
        res[feature + "_N" + suffix] = desc.N
    return res


def sweep_fit_model_trend_exp(
    x: Union[float, np.ndarray], vinf: float, b: float, tau: float
) -> np.ndarray:
    """Sweep model with a linear trend and an exponential decay."""
    return vinf * x + b * (1.0 - np.exp(-x / tau))


def sweep_fit_model_trend_exp_jac(
    x: Union[float, np.ndarray], vinf: float, b: float, tau: float
) -> np.ndarray:
    """Jacobian of the sweep model with a linear trend and an exponential decay.

    :return: matrix with columns being derivatives w.r.t. parameters
    """
    jac = np.zeros((x.size, 3))
    jac[:, 0] = x
    jac[:, 1] = 1.0 - np.exp(-x / tau)
    jac[:, 2] = -b * x * np.exp(-x / tau) / tau**2
    return jac


def sweep_fit_model_exp(
    x: Union[float, np.ndarray], vinf: float, b: float, tau: float
) -> np.ndarray:
    """Sweep model with an exponential decay."""
    return b * (1.0 - np.exp(-x / tau))


def sweep_fit_model_exp_jac(
    x: Union[float, np.ndarray], vinf: float, b: float, tau: float
) -> np.ndarray:
    """Jacobian of the sweep model with an exponential decay.

    :return: matrix with columns being derivatives w.r.t. parameters
    """
    jac = np.zeros((x.size, 3))
    jac[:, 0] = 0.0
    jac[:, 1] = 1.0 - np.exp(-x / tau)
    jac[:, 2] = -b * x * np.exp(-x / tau) / tau**2
    return jac


def fit_sweeps(
    simulation: Simulation,
    sim_num: int,
    coord_cols: List[str],
    plot: bool = False,
    ini_tau: float = 50e-6,
    upper_tau_boundary: float = 350e-6,
    fit_model: Callable = sweep_fit_model_trend_exp,
    fit_model_jac: Callable = sweep_fit_model_trend_exp_jac,
) -> pd.DataFrame:
    """Fits sweeps in the simulation with a given function."""

    sim_obj = simulation.shift_features(coord_cols)
    sim_data = pd.DataFrame(sim_obj.data)
    params = SimulationParameters.from_config(sim_obj.config)

    G_0 = 3.0
    alpha = params.gammaBeadR / params.gammaMT
    v_free = params.F / params.gamma_dmbl
    ini_vinf = v_free * np.exp(params.G / G_0)
    iniB = ini_tau * params.F / params.gammaBeadR * (alpha + 1) / (2 * alpha + 1)

    x_scaling_facor = 1e-3
    y_scaling_factor = 0.2
    norm_vec = np.array(
        [x_scaling_facor / y_scaling_factor, 1 / y_scaling_factor, 1 / x_scaling_facor]
    )
    ini_vec = np.array([ini_vinf, iniB, ini_tau]) * norm_vec
    par_num = len(ini_vec)
    bounds = (np.zeros(3), np.array([np.inf, np.inf, upper_tau_boundary]) * norm_vec)

    def fun(pars, x, y):
        return fit_model(x, *pars) - y

    def fun_jac(pars, x, y):
        return fit_model_jac(x, *pars)

    sweep_boundaries = sim_data.index[
        sim_data["direction"].diff().fillna(1) != 0
    ].to_list() + [sim_data.shape[0]]
    sweeps = list(get_consecutive_pairs(sweep_boundaries))

    res = []
    for i, sweep in enumerate(sweeps):
        sweep_data = sim_data[
            (sim_data.index >= sweep[0]) & (sim_data.index < sweep[1])
        ]

        x, y = sweep_data["Time"].to_numpy(), sweep_data["xBeadr"].to_numpy()
        x = (x - x[0]) / x_scaling_facor
        y = (y - y[0]) / y_scaling_factor

        direction = sim_data.at[sweep[0], "direction"]
        if direction < 0:
            y *= -1

        max_trials = 5
        trials = 0
        not_converged = True
        perturb = np.ones(3)
        while not_converged and trials < max_trials:
            try:
                res_lsq = least_squares(
                    fun,
                    x0=ini_vec * perturb,
                    args=(x, y),
                    bounds=bounds,
                    jac=fun_jac,
                    max_nfev=50000,
                )

                ysize = len(res_lsq.fun)
                cost = 2 * res_lsq.cost  # res.cost is half sum of squares!
                popt = res_lsq.x

                # Do Moore-Penrose inverse discarding zero singular values.
                _, s, VT = svd(res_lsq.jac, full_matrices=False)
            except (LinAlgError, ValueError):

                def generate_perturb():
                    return (np.random.rand(1) * 0.01 - 0.005) + 1

                perturb = generate_perturb()
                while any(ini_vec * perturb > bounds[1]) or any(
                    ini_vec * perturb < bounds[0]
                ):
                    perturb = generate_perturb()
                trials += 1
            else:
                not_converged = False
        if not_converged:
            print(f"Failed to fit {sim_num} due to SVD, skipping")
            return None

        threshold = np.finfo(float).eps * max(res_lsq.jac.shape) * s[0]
        s = s[s > threshold]
        VT = VT[: s.size]
        pcov = np.dot(VT.T / s**2, VT)

        if pcov is None:  # indeterminate covariance
            pcov = np.zeros((len(popt), len(popt)), dtype=float)
            pcov.fill(np.nan)
        else:
            if ysize > par_num:
                s_sq = cost / (ysize - par_num)
                pcov = pcov * s_sq
            else:
                pcov.fill(np.nan)

        popt /= norm_vec
        pcov /= norm_vec**2
        x *= x_scaling_facor
        y *= y_scaling_factor

        ypred = fit_model(x, *popt)
        rsquare = r2_score(y, ypred)
        rsquared_adj = 1 - (1 - rsquare) * (len(x) - 1) / (len(x) - 1 - par_num)

        if plot:
            fig, ax = plt.subplots()
            ax.scatter(x, y, s=4)
            ax.plot(x, ypred, ls="--", c="r")
            title = ", ".join(
                f"{var} = {eng_str(fit, mant_prec=1)}"
                for var, fit in zip(["v_inf", "B", "tau"], popt)
            )
            ax.set_title(title)
            fin_plot()

        res.append(
            np.concatenate(
                [
                    [i, direction],
                    [sim_num, params.G, params.F],
                    popt,
                    np.sqrt(np.diag(pcov)),
                    [rsquared_adj, len(x), ((ypred - y) ** 2).sum() / len(x)],
                ]
            )
        )

    fit_data = pd.DataFrame(
        np.vstack(res),
        columns=[
            "sweep",
            "direction",
            "exp_num",
            "G",
            "F",
            "vinf",
            "b",
            "tau",
            "vinf_sd",
            "b_sd",
            "tau_sd",
            "r2_adj",
            "length",
            "tss",
        ],
    )
    return fit_data


def fit_individual_sweep(
    sweeps: Dict[str, pd.Series],
    F: float,
    plot: bool = False,
    ini_tau: float = 50e-6,
    upper_tau_boundary: float = 350e-6,
    fit_model: Callable = sweep_fit_model_trend_exp,
    fit_model_jac: Callable = sweep_fit_model_trend_exp_jac,
) -> pd.DataFrame:
    """Fits given sweeps with a given function."""

    gamma_dmbl = 0.03
    gammaBeadR = 0.0052

    alpha = gammaBeadR / (gamma_dmbl - gammaBeadR)
    v_free = F / gamma_dmbl
    ini_vinf = v_free / 3
    iniB = ini_tau * F / gammaBeadR * (alpha + 1) / (2 * alpha + 1)

    x_scaling_facor = 1e-3
    y_scaling_factor = 0.2
    norm_vec = np.array(
        [x_scaling_facor / y_scaling_factor, 1 / y_scaling_factor, 1 / x_scaling_facor]
    )
    ini_vec = np.array([ini_vinf, iniB, ini_tau]) * norm_vec
    par_num = len(ini_vec)
    bounds = (np.zeros(3), np.array([np.inf, np.inf, upper_tau_boundary]) * norm_vec)

    def fun(pars, x, y):
        return fit_model(x, *pars) - y

    def fun_jac(pars, x, y):
        return fit_model_jac(x, *pars)

    res = []
    for i, (sim_num, sweep) in enumerate(sweeps.items()):
        x, y = sweep.index.to_numpy(), sweep.to_numpy()
        x = (x - x[0]) / x_scaling_facor
        y = (y - y[0]) / y_scaling_factor

        max_trials = 5
        trials = 0
        not_converged = True
        perturb = np.ones(3)
        while not_converged and trials < max_trials:
            try:
                res_lsq = least_squares(
                    fun,
                    x0=ini_vec * perturb,
                    args=(x, y),
                    bounds=bounds,
                    jac=fun_jac,
                    max_nfev=50000,
                )

                ysize = len(res_lsq.fun)
                cost = 2 * res_lsq.cost  # res.cost is half sum of squares!
                popt = res_lsq.x

                # Do Moore-Penrose inverse discarding zero singular values.
                _, s, VT = svd(res_lsq.jac, full_matrices=False)
            except (LinAlgError, ValueError):

                def generate_perturb():
                    return (np.random.rand(1) * 0.01 - 0.005) + 1

                perturb = generate_perturb()
                while any(ini_vec * perturb > bounds[1]) or any(
                    ini_vec * perturb < bounds[0]
                ):
                    perturb = generate_perturb()
                trials += 1
            else:
                not_converged = False
        if not_converged:
            print(f"Failed to fit sweep {i} due to SVD, skipping")
            return None

        threshold = np.finfo(float).eps * max(res_lsq.jac.shape) * s[0]
        s = s[s > threshold]
        VT = VT[: s.size]
        pcov = np.dot(VT.T / s**2, VT)

        if pcov is None:  # indeterminate covariance
            pcov = np.zeros((len(popt), len(popt)), dtype=float)
            pcov.fill(np.nan)
        else:
            if ysize > par_num:
                s_sq = cost / (ysize - par_num)
                pcov = pcov * s_sq
            else:
                pcov.fill(np.nan)

        popt /= norm_vec
        pcov /= norm_vec**2
        x *= x_scaling_facor
        y *= y_scaling_factor

        ypred = fit_model(x, *popt)
        rsquare = r2_score(y, ypred)
        rsquared_adj = 1 - (1 - rsquare) * (len(x) - 1) / (len(x) - 1 - par_num)

        if plot:
            fig, ax = plt.subplots()
            ax.scatter(x, y, s=4)
            ax.plot(x, ypred, ls="--", c="r")
            title = ", ".join(
                f"{var} = {eng_str(fit, mant_prec=1)}"
                for var, fit in zip(["v_inf", "B", "tau"], popt)
            )
            ax.set_title(title)
            fin_plot()

        res.append(
            np.concatenate(
                [
                    [i, 1],
                    [sim_num, np.nan, F],
                    popt,
                    np.sqrt(np.diag(pcov)),
                    [rsquared_adj, len(x), ((ypred - y) ** 2).sum() / len(x)],
                ]
            )
        )
    fit_data = pd.DataFrame(
        np.vstack(res),
        columns=[
            "sweep",
            "direction",
            "exp_num",
            "G",
            "F",
            "vinf",
            "b",
            "tau",
            "vinf_sd",
            "b_sd",
            "tau_sd",
            "r2_adj",
            "length",
            "tss",
        ],
    )
    return fit_data


def process_fit_data(data: pd.DataFrame, analysis: Analysis) -> pd.DataFrame:
    """Processes simulation sweeps fit data into a DataFrame in wide format.

    :param data: sweeps fit data as returned by ``fit_individual_sweep`` or
        ``fit_sweeps`` functions
    :param analysis: corresponding Analysis instance
    :return: a DataFrame with the simulation results in wide format
    """
    duration_threshold = 30  # points
    cutoff = 1.5  # relaxation time
    critical_num_points = 18  # Keep not less than critNum of points, was 8 !!!! (long, long, long time ago)
    time_between_frames = 15e-6  # s

    cutoff_points = (np.ceil(data["tau"] * cutoff / time_between_frames)).astype(int)

    # How many points to cut off the beginning of each sweep
    data["time_offsets"] = np.minimum(
        np.clip(data["length"] - critical_num_points, 0, None), cutoff_points
    )

    data.sort_values(["exp_num", "sweep"], inplace=True)
    data["sweep_end"] = data.groupby("exp_num", sort=True)["length"].cumsum()
    data["sweep_start"] = data.groupby("exp_num")["sweep_end"].shift(1, fill_value=0)
    data["sweep_end"] -= 1

    data = convert_col_to_types(
        data,
        {
            "int": [
                "exp_num",
                "sweep_start",
                "sweep_end",
                "time_offsets",
                "length",
                "direction",
                "sweep",
            ]
        },
    )

    data["sweep_start"] += data["time_offsets"]

    results = {}

    def process_sim(exp_num, df):
        df = df[df["length"] >= duration_threshold]

        mask_R = df["direction"] > 0
        sweeps_plus = expand_ranges(
            df.loc[mask_R, ["sweep_start", "sweep_end"]].to_records(index=False)
        )
        sweeps_minus = expand_ranges(
            df.loc[~mask_R, ["sweep_start", "sweep_end"]].to_records(index=False)
        )

        simulation = deepcopy(analysis.sims[exp_num])
        sim_params = SimulationParameters.from_config(simulation.config)

        simulation["Ftot"] = (
            simulation["xTrapr"] - simulation["xBeadr"]
        ) * sim_params.k2 + (
            simulation["xTrapl"] - simulation["xBeadl"]
        ) * sim_params.k1

        features_rename_dict = {
            "deltaG": "avg_dG",
            "Ftot": "Ftot",
            "logpotentialForce": "Fmol",
        }

        # TODO use proper flag; another type of Analysis? Like uffc_single or uffc_multiple
        one_molecule = "deltaG" in simulation
        if one_molecule:
            simulation["deltaG"] /= kT
        else:
            simulation["deltaG"] = simulation["deltaG1"] / kT
            simulation["deltaG2"] /= kT
            features_rename_dict["deltaG2"] = "avg_dG_2"

            simulation["logpotentialForce"] = (
                simulation["logpotentialForce1"] + simulation["logpotentialForce2"]
            )

        results[exp_num] = sim_params.as_dict()
        results[exp_num]["sweeps_R"] = df.loc[mask_R, "sweep"].max()
        results[exp_num]["sweeps_L"] = df.loc[~mask_R, "sweep"].max()

        plus_end_stats, minus_end_stats = {}, {}
        for res_dict, idx, s in zip(
            [plus_end_stats, minus_end_stats], [sweeps_plus, sweeps_minus], ["_R", "_L"]
        ):
            for feature, feature_name in features_rename_dict.items():
                res_dict[feature_name] = describe_array(simulation[feature][idx])
            results[exp_num].update(descriptions_to_record(res_dict, suffix=s))

    Parallel(n_jobs=6, require="sharedmem")(
        delayed(process_sim)(exp_num, df)
        for exp_num, df in tqdm(data.groupby("exp_num"))
    )

    res_df = pd.DataFrame.from_records(results).T
    res_df.rename(
        columns={col: swap_dir_prop(col) for col in res_df.columns if "avg_dG" in col}
    )

    res_df = set_negated_cols(res_df, default_cols_to_negate)
    res_df = set_abs_cols(
        res_df, (set(default_abs_cols) | {"Ftot_R", "Ftot_L"}) - {"Ftot"}
    )

    exp_groupper = data.groupby(["exp_num", "direction"])
    vels = pd.DataFrame(
        {
            "v": exp_groupper["vinf"].mean(),
            "v_sd": exp_groupper["vinf_sd"].apply(lambda x: np.sqrt((x**2).sum())),
            "v_N": exp_groupper.size(),
        }
    ).unstack()

    def formatter(x):
        name, direction = x
        direction = "R" if float(direction) > 0 else "L"
        feat, sep, prop = name.partition("_")
        return f"{feat}_{direction}{sep}{prop}"

    vels.columns = vels.columns.map(formatter)

    return pd.merge(vels, res_df, on="exp_num", how="outer")
