"""Module with various utils."""

import collections.abc as abc
import itertools as it

from dataclasses import asdict
from typing import (
    Any,
    Callable,
    Collection,
    Container,
    Dict,
    Hashable,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import numpy as np
import pandas as pd
import numba as nb

from numba import float64, int32, jit, njit
from scipy.linalg import svd
from scipy.optimize import OptimizeResult, fsolve
from scipy.special import expm1, lambertw, logsumexp, poch
from scipy.stats import multivariate_normal, norm, rankdata
from sklearn.linear_model import LinearRegression
from sklearn.neighbors import NearestNeighbors


class DataclassMixin:
    """Adds `as_dict` method to a dataclass."""

    def as_dict(self):
        return asdict(self)


def inverse_dict_bijection(dct: Dict) -> Dict:
    """Inverse dict assuming bijective relationship.

    :param dct: input dict
    :return: inverted dict
    """
    return {val: key for key, val in dct.items()}


def last_from_iter(iterator: Iterable, default_value: Optional[Any] = None):
    """Get the last value of an iterator.

    :param iterator: input iterator to exhaust
    :param default_value: default value to return if iterator is empty, defaults to None
    :return: last item of the input iterator
    """
    last = default_value
    for last in iterator:
        pass
    return last


def is_iterable_container(el) -> bool:
    """Check, if this object is iterable. dict is not an iterable in this function.

    :param el: element to check
    """
    return isinstance(el, abc.Iterable) and not isinstance(el, (str, bytes, dict))


def flatten(lst: Iterable):
    """Flatten nested iterables to one level.

    :param lst: nested iterable
    :return: iterator over flatten collection
    """
    for el in lst:
        if is_iterable_container(el):
            yield from flatten(el)
        else:
            yield el


def hoeffding_d(
    x: Union[Collection[float], np.ndarray], y: Union[Collection[float], np.ndarray]
) -> float:
    """Calculate Hoeffding dependence measure between two vectors of the same length.

    :param x: first data vector
    :param y: second data vector
    :return: Hoeffding D value
    """
    n = len(x)
    if n != len(y):
        raise ValueError(f"Lengths of arrays are unequal: {n} and {len(y)}")
    if n < 5:
        raise ValueError(f"Length of input array must be at least 5, got {n}")

    x = np.asarray(x)
    y = np.asarray(y)

    r = rankdata(x)
    s = rankdata(y)

    c = np.zeros_like(r, dtype=float)
    for i, ri in enumerate(r):
        si = s[i]
        rless = r < ri
        sless = s < si

        req = r == ri
        seq = s == si

        res = (rless & sless).sum()
        res += 0.25 * ((req & seq).sum() - 1)
        res += 0.5 * ((req & sless).sum() + (rless & seq).sum())

        c[i] = res

    q_sum = np.sum((r - 1) * (r - 2) * (s - 1) * (s - 2))
    s_sum = np.sum(c * (c - 1))
    r_sum = np.sum(c * (r - 2) * (s - 2))

    return (
        30 * (-2 * (n - 2) * r_sum + (n - 3) * (n - 2) * s_sum + q_sum) / poch(n - 4, 5)
    )


def hausdorff(
    x_a: np.ndarray, x_b: np.ndarray, distance_function: Callable[..., float]
) -> float:
    """Compute Hausdorff distance between two sets of points (A and B) according to specified distance function
    From https://github.com/mavillan/py-hausdorff/

    :param x_a: points from set A
    :param x_b: points from set B
    :param distance_function: distance function to use
    :return: Hausdorff distance between sets A and B
    """
    n_a = x_a.shape[0]
    n_b = x_b.shape[0]
    cmax = 0.0
    for i in range(n_a):
        cmin = np.inf
        for j in range(n_b):
            d = distance_function(x_a[i, :], x_b[j, :])
            if d < cmin:
                cmin = d
            if cmin < cmax:
                break
        if cmax < cmin < np.inf:
            cmax = cmin

    for j in range(n_b):
        cmin = np.inf
        for i in range(n_a):
            d = distance_function(x_a[i, :], x_b[j, :])
            if d < cmin:
                cmin = d
            if cmin < cmax:
                break
        if cmax < cmin < np.inf:
            cmax = cmin
    return cmax


def to_narrow(df: pd.DataFrame, side_labels: tuple = ("R", "L")) -> pd.DataFrame:
    """Transform DataFrame with results from wide format (right and left features in one row) to narrow
    format with direction labels.

    :param df: input DataFrame to convert
    :param side_labels: tuple of labels for motion to the right and to the left
    :return: data in narrow format
    """
    side_cols = {
        side_label: {
            col: col.replace(f"_{side_label}", "")
            for col in df.columns
            if f"_{side_label}_" in col or col.endswith(f"_{side_label}")
        }
        for side_label in side_labels
    }
    side_cols_list = [vv for v in side_cols.values() for vv in v.keys()]
    common_cols = [col for col in df.columns if col not in side_cols_list]

    res = pd.DataFrame()
    for side in side_labels:
        df_side = df[common_cols + list(side_cols[side].keys())].rename(
            columns=side_cols[side]
        )
        df_side["direction"] = side
        res = pd.concat([res, df_side], axis=0, sort=False)

    return res


def add_rates(ndf: pd.DataFrame, rdf: pd.DataFrame) -> pd.DataFrame:
    """Add rates info to results in narrow format.

    :param rdf: DataFrame with rates
    :param ndf: results DataFrame in narrow format
    """

    return ndf.merge(rdf, how="left", on=["exp_num", "direction"])


def prepare_nonlinear_cols(
    df: pd.DataFrame, x_col: str = "Fmol", y_col: str = "v"
) -> None:
    """NOT USED. WARNING: CHECK THE IMPLEMENTATION

    TODO refactor

    :param df:
    :param x_col:
    :param y_col:
    :return:
    """
    df.loc[:, "log_ratio"] = np.log(df[y_col] / df[x_col])
    complex_lambert = lambertw(
        np.sqrt(
            #         df[x_col] ** 2
            +df[y_col]
            ** 2
        )
    )
    if (np.imag(complex_lambert) != 0).any():
        raise RuntimeError("Lambert W yielded complex results!")
    df.loc[:, "lambert"] = np.real(complex_lambert)

    x = df["lambert"][df["direction"] == "R"].values
    y = df["log_ratio"][df["direction"] == "R"].values
    reg = LinearRegression().fit(x.reshape(-1, 1), y)
    df.loc[:, "detrended"] = df["log_ratio"] - df["lambert"] * reg.coef_


def cartesian2polar(x: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Convert cartesian coordinates into polar.

    :param x: X values array
    :param y: Y values array
    :return: tuple of polar angles and radii arrays
    """
    r = np.sqrt(x**2 + y**2)
    theta = np.arctan(y / x)
    return theta, r


def add_asymmetry_narrow(ndf: pd.DataFrame, col: str) -> pd.DataFrame:
    """Add asymmetry metric to narrow DataFrame with data.

    :param ndf: data in narrow format
    :param col: column identifying equivalence class (e.g. 'exp_name')
    :return: data with `assym` column added
    """
    tmp_df = ndf.set_index(col)
    mask = tmp_df["direction"] == "R"

    return pd.merge(
        ndf,
        (tmp_df["v"][mask] / tmp_df["v"][~mask]).rename("assym"),
        left_on=col,
        right_index=True,
        how="left",
    )


def get_contigious_ranges(arr: Sequence[int]) -> List[Tuple[int, int]]:
    """Extract contiguous ranges from 1D sequence of integers.

    :param arr: 1D sequence of numbers
    :return: list of tuples representing contiguous ranges
    """
    res = []
    for k, g in it.groupby(enumerate(sorted(arr)), lambda t: t[0] - t[1]):
        nums = map(lambda x: x[1], g)
        res.append((next(nums), last_from_iter(nums)))
    return res


def get_contigious_ranges_tuples(
    arr: Sequence[Tuple[int, int]]
) -> List[Tuple[int, int]]:
    """Extract contiguous ranges from sequence of ranges.

    :param arr: sequence of ranges as 2-tuples
    :return: list of tuples representing contiguous ranges
    """
    arr = sorted(arr)
    if len(arr) == 1:
        return arr

    arr_fwd = iter(arr)
    curr_rng = list(next(arr_fwd))

    res = []
    for x, y in zip(arr, arr_fwd):
        if y[0] - x[1] == 1:
            curr_rng[1] = y[1]
        else:
            res.append(tuple(curr_rng))
            curr_rng = list(y)
    res.append(tuple(curr_rng))

    return res


def round_from_zero(x: float) -> float:
    """Rounds input float x away from zero."""
    return np.fix(x) + np.sign(x)


def calc_experimental_like(
    force: float,
    velocity: float,
    vd_data: pd.DataFrame,
    force_col: str = "Fmol",
    error_force: float = 0.5,
    error_velocity: float = 10.0,
) -> float:
    """Calculates log-likelihood of the force-velocity observation given experimental data.

    :param force: observed force value
    :param velocity: observed velocity value
    :param vd_data: experimental data to use
    :param force_col: column with force values
    :param error_force: uncertainty in force value
    :param error_velocity: uncertainty in velocity value
    :return: log-likelihood (base 10) of the point given the exp data
    """
    num_arr, denom_arr = np.empty(len(vd_data)), np.empty(len(vd_data))

    for i, (x, y) in enumerate(vd_data[[force_col, "v"]].abs().itertuples(index=False)):
        num_arr[i] = multivariate_normal.logpdf(
            [force, velocity], [x, y], [[error_force**2, 0], [0, error_velocity**2]]
        )
        denom_arr[i] = norm.logpdf(force, loc=x, scale=error_force)
    loglike = (logsumexp(num_arr) - logsumexp(denom_arr)) / np.log(10.0)
    return loglike


def calc_experimental_like_many(
    df: pd.DataFrame,
    vd_data: pd.DataFrame,
    force_col: str = "Fmol",
    log: bool = True,
    norm: bool = False,
) -> float:
    """Calculates (log-)likelihood of the simulation data given experimental data.

    :param df: simulation data to use
    :param vd_data: experimental data to use
    :param force_col: column with force values
    :param log: if True, return log-likelihood (base 10)
    :param norm: if True, return value corresponding to average (log-)likelihood per observation
    :return:
    """
    loglike = 0.0
    for x, y in df.abs().itertuples(index=False):
        loglike += calc_experimental_like(x, y, vd_data, force_col)
    if norm:
        if (denom := len(df)) != 0:
            loglike /= denom
        else:
            return np.nan

    if not np.isfinite(loglike):
        return np.nan

    if log:
        return loglike
    return 10**loglike


def get_family_like(
    ndf: pd.DataFrame,
    exp_data: pd.DataFrame,
    keys: Optional[Collection[str]],
    force_col: str = "Fmol",
    log: bool = True,
    norm: bool = False,
) -> pd.DataFrame:
    """

    :param ndf: simulation data
    :param exp_data: experimental data to compare with
    :param force_col: column with force values
    :param log: if True, use log-metric, otherwise use metric
    :param norm: if True, return value corresponding to average log-likelihood per observation
    :param keys: keys that define a family of pints for likelihood computation
    :return: DataFrame with likelihood values
    """
    if keys is None:
        keys = ["G", "G_rot", "B"]
    ndf_filtered = (
        ndf.dropna(subset=[force_col, "v"]).groupby(keys).filter(lambda x: len(x) > 1)
    )
    return (
        ndf_filtered.groupby(keys)[[force_col, "v"]]
        .apply(calc_experimental_like_many, exp_data, force_col, log, norm)
        .rename("like")
        .reset_index()
    )


def get_family_ss_nearest(
    ndf: pd.DataFrame,
    exp_data: pd.DataFrame,
    keys: Optional[Collection[str]],
    force_col: str = "Fmol",
    log: bool = True,
    norm: bool = False,
):
    if keys is None:
        keys = ["G", "G_rot", "B"]

    nbrs = NearestNeighbors(
        n_neighbors=1, metric="seuclidean", metric_params={"V": [0.25, 100]}
    ).fit(exp_data[[force_col, "v"]].values)
    if norm:
        _agg_func = np.mean
    else:
        _agg_func = np.sum

    ndf_filtered = ndf.groupby(keys).filter(lambda x: len(x) > 1)
    like_res = ndf_filtered.groupby(keys).apply(
        lambda x: _agg_func(nbrs.kneighbors(x[[force_col, "v"]].values)[0] ** 2)
    )
    if log:
        like_res = -np.log10(like_res)

    like_res = like_res.rename("like").reset_index()

    return like_res


def pareto_front(points, return_mask: bool = True):
    """Find the pareto-efficient points.

    Modified code from https://stackoverflow.com/a/40239615

    :param points: (n_points, dimension) array
    :param return_mask: whether to return a mask for Pareto front or indices of Pareto front elements
    :return: array of indices of pareto-efficient points or corresponding boolean mask
    """
    is_efficient = np.arange(points.shape[0])
    n_points = points.shape[0]
    next_point_index = 0  # Next index in the is_efficient array to search for
    while next_point_index < len(points):
        nondominated_point_mask = np.any(points < points[next_point_index], axis=1)
        nondominated_point_mask[next_point_index] = True
        is_efficient = is_efficient[nondominated_point_mask]  # Remove dominated points
        points = points[nondominated_point_mask]
        next_point_index = np.sum(nondominated_point_mask[:next_point_index]) + 1
    if return_mask:
        is_efficient_mask = np.full(n_points, False)
        is_efficient_mask[is_efficient] = True
        return is_efficient_mask
    else:
        return is_efficient


def eng_str(x: float, sep: str = "e", mant_prec: Optional[int] = None) -> str:
    exponent = np.log10(abs(x))
    exponent -= exponent % 3
    mantissa = x / 10**exponent

    mant_str = f"{mantissa}" if mant_prec is None else f"{mantissa:.{mant_prec}f}"
    if exponent == 0:
        return mant_str
    return f"{mant_str}{sep}{exponent:.0f}"


def get_consecutive_pairs(lst: Iterable) -> Iterator:
    _iter = iter(lst)
    next(_iter)
    return zip(lst, _iter)


def inverse_dict_of_lists(
    dct: Dict[Hashable, List[Hashable]]
) -> Dict[Hashable, Hashable]:
    list_vals = set()
    res = dict()
    for key, val in dct.items():
        if not list_vals.isdisjoint(val):
            raise ValueError("Lists in dct.values() should not intersect")
        list_vals.update(val)
        for el in val:
            res[el] = key
    return res


def convert_col_to_types(
    df: pd.DataFrame, type2cols: Dict[Union[str, Type, np.dtype], List[str]]
) -> pd.DataFrame:
    return df.astype(inverse_dict_of_lists(type2cols))


def expand_ranges(
    range_list: Iterable[Tuple[int, int]], right_inclusive: bool = True
) -> List[int]:
    """Expand iterable of ranges as tuples into flat list of numbers."""
    return list(
        flatten(
            list(range(start, end + 1 if right_inclusive else end))
            for start, end in range_list
        )
    )


def swap_dir_prop(x: str) -> str:
    """Swap direction and property levels in feature name.

    'feature_name_L_sd' -> 'feature_name_sd_L'
    'feature_name_R' -> 'feature_name_R'

    :param x: string '<feature name>_[RL]_<property>' or '<feature name>_[RL]'
    :return: string '<feature name>_<property>_[RL]' or '<feature name>_[RL]'
    """
    if x[-2:] in {"_R", "_L"}:
        return x
    suff, directory, prop = x.rsplit("_", maxsplit=2)
    return f"{suff}_{prop}_{directory}"


def composite_grid(
    n: int, m: float = 0.75, ratio: float = 0.5, a: float = 0.0, b: float = 1.0
) -> np.ndarray:
    """Create composite linear-geometric n-grid for [a, b] interval.

    First k = floor(n * ratio) points are evenly spaced in [a, m] interval, last n - k points
    constitute a geometric progression in [m, b] interval.

    Input parameters must satisfy the following conditions:
    - 0 < m <= 1
    - 0 < ratio <= 1
    - b >= a

    :param n: number of points in grid. If odd, increased by 1
    :param m: for [a, m] use equidistant grid, for [m, b] use geometrically decreasing intervals.
    :param ratio: allocate this fraction of grid points to equidistant [a, m] grid.
    :param a: left boundary of interval
    :param b: right boundary of interval (inclusive)
    :return: array of grid points
    """
    if not (0 < m <= 1 and 0 < ratio <= 1 and a < b):
        raise ValueError(
            "Wrong interval or grid parameters. See doc string for constraints."
        )
    k = int(np.floor(n * ratio))
    linear = np.linspace(0, m, k)

    def eqn(le):
        q = -expm1(le)
        t = np.log(q) * (n - k)
        return -k * np.exp(le) / expm1(t) - m / (1 - m)

    le = fsolve(eqn, np.array([-1.0]), full_output=False, factor=0.1, epsfcn=1e-8)[0]
    lq = np.log(-expm1(le))

    lnk = np.log(-expm1(lq * (n - k)))
    log_prog = lq * np.arange(1, n - k + 1, 1, dtype=int) - lnk

    offset = np.exp(np.log(-expm1(lq * (n - k) + np.log(m))) - lnk)
    geom = offset + (m - 1) * np.exp(log_prog)

    if (eps := np.exp(le)) < 1e-3:
        print(
            f"Ratio for geometric intervals is small, 1 - q = {eps:.3e}, consider using other parameters."
        )

    return np.hstack([linear, geom]) * (b - a) + a


def extend_grid_with_tolerance(
    arr: Union[np.ndarray, Collection],
    tol: float,
    a: Optional[float] = None,
    b: Optional[float] = None,
) -> Tuple[List[float], List[int]]:
    """Add points to sorted arr, capped with a and b, so that difference between points is less than tol.

    :param arr: original array with values to extend
    :param tol: maximal distance between two consecutive points in the resulting grid
    :param a: optional value to start grid from (left boundary)
    :param b: optional value to end grid with (right boundary)
    :return: tuple of two lists:
    - list of grid values
    - list of indices for original array values in the grid
    """
    iter_base = iter(arr)
    if a is not None:
        iter_base = it.chain([a], iter_base)
    if b is not None:
        iter_base = it.chain(iter_base, [b])

    iter_base, iter_next = it.tee(iter_base)
    next(iter_next)

    res = []
    orig_data_idx = []
    for x_c, x_n in zip(iter_base, iter_next):
        diff = x_n - x_c
        n_points = int(np.ceil(diff / tol))
        if len(res) > 0 or a is None:
            orig_data_idx.append(len(res))
        res.extend(x_c + diff * i / n_points for i in range(n_points))
    res.append(x_n)
    if b is None:
        orig_data_idx.append(len(res) - 1)
    return res, orig_data_idx


@jit(nopython=True)
def downsample_closest(
    arr: np.ndarray, ref_grid: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    curr_idx = 0
    curr_val = ref_grid[curr_idx]

    max_len = ref_grid.shape[0]
    arr_len = arr.shape[0]
    res = np.empty(ref_grid.shape[0], dtype=arr.dtype)
    res_idx = np.empty(ref_grid.shape[0], dtype="int")

    saved = 0

    first_a = arr[0]
    if curr_val <= first_a:
        for curr_idx, curr_val in enumerate(ref_grid):
            if curr_val > first_a:
                break

        if ref_grid[curr_idx - 1] < arr[1] <= curr_val:
            start = curr_idx - 1
        else:
            start = curr_idx

        res[saved] = arr[0]
        res_idx[saved] = 0
        saved += 1

        curr_idx = start + 1
        curr_val = ref_grid[curr_idx]

    i = saved
    while i < arr_len:
        x = arr[i]
        if x < curr_val:
            i += 1
            continue
        prev_x = arr[i - 1]
        if curr_val - prev_x >= x - curr_val:
            res[saved] = x
            res_idx[saved] = i
        else:
            res[saved] = prev_x
            res_idx[saved] = i - 1
            i -= 1
        saved += 1

        x_idx = i
        if i < arr_len - 1:
            x = arr[i + 1]
            x_idx = i + 1

        if curr_idx >= max_len - 1:
            break
        tmp = curr_idx
        for curr_idx in range(tmp + 1, max_len):
            curr_val = ref_grid[curr_idx]
            if curr_val > x:
                if (
                    ref_grid[curr_idx - 1] - res[saved - 1]
                    <= x - ref_grid[curr_idx - 1]
                ):
                    break
                if x - ref_grid[curr_idx - 1] <= curr_val - x:
                    res[saved] = x
                    res_idx[saved] = x_idx
                    saved += 1
                    i += 1
                    curr_idx += 1
                    if curr_idx >= max_len:
                        break
                    curr_val = ref_grid[curr_idx]
                break
        i += 1

    return res[:saved], res_idx[:saved]


def moving_average(x: np.ndarray, w: int) -> np.ndarray:
    """Computes moving average of array with given window width."""
    return np.convolve(x, np.ones(w, dtype=x.dtype), "valid") / w


def derivative(
    x: np.ndarray, y: np.ndarray, m_max: int, x0: float, order: int
) -> np.ndarray:
    """Compute finite difference approximation for derivatives on uneven grid.

    From Fornberg (1988) with minor modifications.

    :param x: input x array
    :param y: input y array
    :param m_max: maximal degree of derivative to compute (inclusive)
    :param x0: point at which to compute derivatives
    :param order: order of approximation to compute. Eventually may become a higher one
    :return: array of derivative values at given point starting from 0 to m_max order
    """
    res = _derivative(x, y, m_max, x0, order)
    if res is None:
        raise RuntimeError(f"n_max = {order + m_max - 1} > N = {len(x) - 1}")
    return res


@njit
def _derivative(
    x: np.ndarray, y: np.ndarray, m_max: int, x0: float, order: int
) -> Optional[np.ndarray]:
    """NJIT-ed internals of derivative function.

    Refer to it for more details on parameters.
    """
    N = len(x) - 1
    d = np.zeros((m_max + 1, N + 1, N + 1))

    new_idx = _reorder_pivoting(x, x0)
    x = x[new_idx].copy()
    y = y[new_idx].copy()

    n_max = order + m_max - 1  # N in original
    if n_max > N:
        return None

    d[0, 0, 0] = 1.0

    c1 = 1.0

    for n in range(1, n_max + 1):
        c2 = 1.0
        m_rng = min(n, m_max)
        for v in range(n):
            c3 = x[n] - x[v]
            c2 = c2 * c3
            for m in range(m_rng + 1):
                d[m, n, v] = (
                    (x[n] - x0) * d[m, n - 1, v] - m * d[m - 1, n - 1, v]
                ) / c3

        for m in range(m_rng + 1):
            d[m, n, n] = (
                c1
                / c2
                * (m * d[m - 1, n - 1, n - 1] - (x[n - 1] - x0) * d[m, n - 1, n - 1])
            )
        c1 = c2

    ders = np.zeros(m_max + 1)
    for m in range(m_max + 1):
        n = order + m - 1
        ders[m] = d[m, n, : n + 1] @ y[: n + 1]

    return ders


@njit(nb.int64[:](nb.float64[:], nb.float64))
def _reorder_pivoting(x, val):
    if val > x[-1]:
        return np.arange(len(x) - 1, -1, -1)
    elif val < x[0]:
        return np.arange(len(x))

    idx = np.searchsorted(x, val)

    seq = [idx]
    i = 1
    inc = 1
    while 0 <= idx + i * inc < len(x):
        seq.append(idx + i * inc)
        inc *= -1
        i += inc > 0

    if idx < len(x) // 2:
        seq.extend(range(idx - i * inc + 1, len(x)))
    else:
        seq.extend(range(idx - i * inc + 1))
    return np.array(seq)


def derivative_array(
    x: np.ndarray, y: np.ndarray, x0_arr: np.ndarray, m_max: int, order: int
) -> np.ndarray:
    """Compute derivatives numerically at given points.

    :param x: input x array
    :param y: input y array
    :param m_max: maximal degree of derivative to compute (inclusive)
    :param x0_arr: array of points at which to compute derivatives
    :param order: order of approximation to compute. Eventually may become a higher one
    :return: array of derivative values at given points starting from 0 to m_max order
    - axis 0: points
    - axis 1: order of the derivative
    """
    return np.vstack(
        [derivative(x=x, y=y, x0=x_curr, m_max=m_max, order=order) for x_curr in x0_arr]
    ).T


def unique_stable(arr: np.ndarray) -> np.ndarray:
    """Return unique values in `arr` preserving their order of appearance in array."""
    unique_vals, ind = np.unique(arr, return_index=True)
    return unique_vals[np.argsort(ind)]


def merge_numpy(
    x: np.ndarray, y: np.ndarray, indices: bool = False
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray, np.ndarray]]:
    """Merges two sorted arrays into a single sorted array.

    :param x: first sorted array
    :param y: second sorted array
    :param indices: whether to return indices of original array elements in merged array
    :return: merged sorted array or tuple (merged_array, (idx_x, idx_y))
    """
    res = np.concatenate((x, y))
    if not indices:
        res.sort(kind="stable")
        return res
    else:
        idx = np.argsort(res)
        arr = np.empty(len(res), dtype=int)

        arr[idx] = np.arange(len(res))
        return res, arr[: len(x)], arr[len(x) :]


def impute_grid(
    grid: np.ndarray, imputor: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Return merged array, indices of the grid and imputed values."""
    res, (idx_grid, idx_imputor) = merge_numpy(grid, imputor, indices=True)
    return res, idx_grid, idx_imputor


def otsu_threshold(counts, bin_centers):
    """Gets threshold for a histogram using Otsu method.

    Implementation: from ``threshold_otsu`` function of scikit-image.
    """

    weight1 = np.cumsum(counts)
    weight2 = np.cumsum(counts[::-1])[::-1]

    mean1 = np.cumsum(counts * bin_centers) / weight1
    mean2 = (np.cumsum((counts * bin_centers)[::-1]) / weight2[::-1])[::-1]
    variance12 = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:]) ** 2

    idx = np.argmax(variance12)
    return bin_centers[idx]


@njit(int32[:, :](float64[:], float64[:]))
def common_indices_sorted(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """Return indices of common elements in two sorted arrays.

    :param x: first sorted array
    :param y: second sorted array
    :return: array with two rows of common elements indices in the input arrays
    """
    flip_flag = False
    if x[0] > y[0]:
        x, y = y, x
        flip_flag = True

    i: int = 0
    j: int = 0
    total: int = 0

    n = len(x)
    m = len(y)

    res = np.empty((min(n, m), 2), dtype=np.int32)
    while i < n and j < m:
        if x[i] == y[j]:
            res[total, :] = (i, j)
            total += 1
            j += 1
        if x[i] > y[j]:
            j += 1
        else:
            i += 1
    if flip_flag:
        ret_val = res[:total, ::-1]
    else:
        ret_val = res[:total, :]
    return ret_val.T


def cov_estimate_from_lsq(res: OptimizeResult, absolute_sigma: bool = False):
    """Estimate covariance matrix for parameters from least_squares optimization.

    Code from https://github.com/scipy/scipy/blob/v1.6.3/scipy/optimize/minpack.py#L532-L834

    :param res: OptimizeResult instance returned by least_squares function
    :param absolute_sigma: whether absolute_sigma flag was passed to least_squares. Defaults to False
    :return: covariance matrix
    """
    if not res.success:
        raise RuntimeError("Optimal parameters not found: " + res.message)

    ysize = len(res.fun)
    cost = 2 * res.cost  # res.cost is half sum of squares!
    popt = res.x
    n_params = len(popt)

    # Do Moore-Penrose inverse discarding zero singular values.
    _, s, VT = svd(res.jac, full_matrices=False)
    threshold = np.finfo(float).eps * max(res.jac.shape) * s[0]
    s = s[s > threshold]
    VT = VT[: s.size]
    pcov = np.dot(VT.T / s**2, VT)

    if pcov is None:
        # indeterminate covariance
        pcov = np.zeros((len(popt), len(popt)), dtype=float)
        pcov.fill(np.inf)
    elif not absolute_sigma:
        if ysize > n_params:
            s_sq = cost / (ysize - n_params)
            pcov = pcov * s_sq
        else:
            pcov.fill(np.inf)
    return pcov


def check_argument(arg_value, arg_name: str, allowed_values: Container) -> None:
    if arg_value not in allowed_values:
        raise ValueError(f"Unknown {arg_name}={arg_value}, allowed {allowed_values}")
